# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
##
# MIT License
#
# Copyright (c) 2019 Masaaki Goshima
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##
import base64
import dataclasses as dc
import datetime
import enum
import glob
import os.path
import typing as ta

from ....lite.check import check
from .ast import AliasYamlNode
from .ast import AnchorYamlNode
from .ast import ArrayYamlNode
from .ast import BoolYamlNode
from .ast import CommentGroupYamlNode
from .ast import FloatYamlNode
from .ast import InfinityYamlNode
from .ast import IntegerYamlNode
from .ast import LiteralYamlNode
from .ast import MapKeyYamlNode
from .ast import MappingKeyYamlNode
from .ast import MappingValueYamlNode
from .ast import MappingYamlNode
from .ast import MapYamlNode
from .ast import MapYamlNodeIter
from .ast import NanYamlNode
from .ast import NullYamlNode
from .ast import SequenceYamlNode
from .ast import StringYamlNode
from .ast import TagYamlNode
from .ast import UnexpectedNodeTypeYamlError
from .ast import YamlFile
from .ast import YamlNode
from .ast import YamlNodeType
from .ast import yaml_sequence_merge_value
from .errors import EofYamlError
from .errors import YamlError
from .errors import YamlErrorOr
from .errors import yaml_error
from .parsing import YAML_PARSE_COMMENTS
from .parsing import YamlOption
from .parsing import YamlParseMode
from .parsing import yaml_allow_duplicate_map_key
from .parsing import yaml_parse_str
from .tokens import YamlReservedTagKeywords
from .tokens import YamlSyntaxError
from .tokens import YamlToken


##


class YamlDecodeContext:
    def __init__(self, values: ta.Optional[ta.Dict[ta.Any, ta.Any]] = None) -> None:
        super().__init__()

        self._values: ta.Dict[ta.Any, ta.Any] = values if values is not None else {}

    def with_value(self, key: ta.Any, value: ta.Any) -> 'YamlDecodeContext':
        return YamlDecodeContext({**self._values, key: value})

    def value(self, key: ta.Any) -> ta.Any:
        return self._values.get(key)

    #

    class _MergeKey:  # noqa
        pass

    def with_merge(self) -> 'YamlDecodeContext':
        return self.with_value(self._MergeKey, True)

    def is_merge(self) -> bool:
        if not isinstance(v := self.value(self._MergeKey), bool):
            return False

        return v

    #

    class _AnchorKey:  # noqa
        pass

    def with_anchor(self, name: str) -> 'YamlDecodeContext':
        anchor_map = self.get_anchor_map()
        new_map: ta.Dict[str, None] = {}
        new_map.update(anchor_map)
        new_map[name] = None
        return self.with_value(self._AnchorKey, new_map)

    def get_anchor_map(self) -> ta.Dict[str, None]:
        if not isinstance(v := self.value(self._AnchorKey), dict):
            return {}

        return v


##


# CommentPosition type of the position for comment.
class YamlCommentPosition(enum.IntEnum):
    HEAD = 0
    LINE = 1
    FOOT = 2


# Comment raw data for comment.
@dc.dataclass()
class YamlComment:
    texts: ta.List[str]
    position: YamlCommentPosition


# LineComment create a one-line comment for CommentMap.
def yaml_line_comment(text: str) -> YamlComment:
    return YamlComment(
        texts=[text],
        position=YamlCommentPosition.LINE,
    )


# HeadComment create a multiline comment for CommentMap.
def yaml_head_comment(*texts: str) -> YamlComment:
    return YamlComment(
        texts=list(texts),
        position=YamlCommentPosition.HEAD,
    )


# FootComment create a multiline comment for CommentMap.
def yaml_foot_comment(*texts: str) -> YamlComment:
    return YamlComment(
        texts=list(texts),
        position=YamlCommentPosition.FOOT,
    )


# CommentMap map of the position of the comment and the comment information.
class YamlCommentMap(ta.Dict[str, ta.List[YamlComment]]):
    pass


##


# MapItem is an item in a MapSlice.
@dc.dataclass()
class YamlMapItem:
    key: ta.Any
    value: ta.Any


# MapSlice encodes and decodes as a YAML map.
# The order of keys is preserved when encoding and decoding.
class YamlMapSlice(ta.List[YamlMapItem]):
    # ToMap convert to map[interface{}]interface{}.
    def to_map(self) -> ta.Dict[ta.Any, ta.Any]:
        return {item.key: item.value for item in self}


##


class YamlBytesReader(ta.Protocol):
    def read(self) -> bytes: ...


class ImmediateYamlBytesReader:
    def __init__(self, bs: bytes) -> None:
        self._bs = bs

    def read(self) -> bytes:
        bs = self._bs
        self._bs = b''
        return bs


##


class YamlDecodeOption(ta.Protocol):
    def __call__(self, d: 'YamlDecoder') -> ta.Optional[YamlError]: ...


##


class YamlDecodeErrors:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    EXCEEDED_MAX_DEPTH = yaml_error('exceeded max depth')


@dc.dataclass()
class DuplicateKeyYamlError(YamlError):
    msg: str
    token: YamlToken

    @property
    def message(self) -> str:
        return self.msg


##


# Decoder reads and decodes YAML values from an input stream.
class YamlDecoder:
    reader: YamlBytesReader
    reference_readers: ta.List[YamlBytesReader]
    anchor_node_map: ta.Dict[str, ta.Optional[YamlNode]]
    anchor_value_map: ta.Dict[str, ta.Any]
    comment_maps: ta.List[YamlCommentMap]
    to_comment_map: ta.Optional[YamlCommentMap] = None
    opts: ta.List[YamlDecodeOption]
    reference_files: ta.List[str]
    reference_dirs: ta.List[str]
    is_recursive_dir: bool = False
    is_resolved_reference: bool = False
    allow_duplicate_map_key: bool = False
    use_ordered_map: bool = False
    parsed_file: ta.Optional[YamlFile] = None
    stream_index: int = 0
    decode_depth: int = 0

    # NewDecoder returns a new decoder that reads from r.
    def __init__(self, r: YamlBytesReader, *opts: YamlDecodeOption) -> None:
        super().__init__()

        self.reader = r
        self.anchor_node_map = {}
        self.anchor_value_map = {}
        self.opts = list(opts)
        self.reference_readers = []
        self.reference_files = []
        self.reference_dirs = []
        self.is_recursive_dir = False
        self.is_resolved_reference = False
        self.allow_duplicate_map_key = False
        self.use_ordered_map = False

        self.comment_maps = []

    MAX_DECODE_DEPTH: ta.ClassVar[int] = 10000

    def step_in(self) -> None:
        self.decode_depth += 1

    def step_out(self) -> None:
        self.decode_depth -= 1

    def is_exceeded_max_depth(self) -> bool:
        return self.decode_depth > self.MAX_DECODE_DEPTH

    def cast_to_float(self, v: ta.Any) -> ta.Any:
        if isinstance(v, float):
            return v
        elif isinstance(v, int):
            return float(v)
        elif isinstance(v, str):
            # if error occurred, return zero value
            try:
                return float(v)
            except ValueError:
                return 0
        return 0

    def map_key_node_to_string(self, ctx: YamlDecodeContext, node: MapKeyYamlNode) -> YamlErrorOr[str]:
        key = self.node_to_value(ctx, node)
        if isinstance(key, YamlError):
            return key
        if key is None:
            return 'null'
        if isinstance(key, str):
            return key
        return str(key)

    def set_to_map_value(
            self,
            ctx: YamlDecodeContext,
            node: YamlNode,
            m: ta.Dict[str, ta.Any],
    ) -> ta.Optional[YamlError]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            self.set_path_comment_map(node)
            if isinstance(n := node, MappingValueYamlNode):
                if n.key.is_merge_key():
                    value = self.get_map_node(n.value, True)
                    if isinstance(value, YamlError):
                        return value

                    it = value.map_range()
                    while it.next():
                        if (err := self.set_to_map_value(ctx, it.key_value(), m)) is not None:
                            return err

                else:
                    key = self.map_key_node_to_string(ctx, n.key)
                    if isinstance(key, YamlError):
                        return key

                    v = self.node_to_value(ctx, n.value)
                    if isinstance(v, YamlError):
                        return v

                    m[key] = v

            elif isinstance(n, MappingYamlNode):
                for value2 in n.values:
                    if (err := self.set_to_map_value(ctx, value2, m)) is not None:
                        return err

            elif isinstance(n, AnchorYamlNode):
                anchor_name = check.not_none(check.not_none(n.name).get_token()).value
                self.anchor_node_map[anchor_name] = n.value

            return None

        finally:
            self.step_out()

    def set_to_ordered_map_value(
            self,
            ctx: YamlDecodeContext,
            node: YamlNode,
            m: YamlMapSlice,
    ) -> ta.Optional[YamlError]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            self.set_path_comment_map(node)
            if isinstance(n := node, MappingValueYamlNode):
                if n.key.is_merge_key():
                    value = self.get_map_node(n.value, True)
                    if isinstance(value, YamlError):
                        return value

                    it = value.map_range()
                    while it.next():
                        if (err := self.set_to_ordered_map_value(ctx, it.key_value(), m)) is not None:
                            return err

                else:
                    key = self.map_key_node_to_string(ctx, n.key)
                    if isinstance(key, YamlError):
                        return key

                    value = self.node_to_value(ctx, n.value)
                    if isinstance(value, YamlError):
                        return value

                    m.append(YamlMapItem(key, value))

            elif isinstance(n, MappingYamlNode):
                for value2 in n.values:
                    if (err := self.set_to_ordered_map_value(ctx, value2, m)) is not None:
                        return err

            return None

        finally:
            self.step_out()

    def set_path_comment_map(self, node: ta.Optional[YamlNode]) -> None:
        if node is None:
            return

        if self.to_comment_map is None:
            return

        self.add_head_or_line_comment_to_map(node)
        self.add_foot_comment_to_map(node)

    def add_head_or_line_comment_to_map(self, node: YamlNode) -> None:
        if isinstance(node, SequenceYamlNode):
            self.add_sequence_node_comment_to_map(node)
            return

        comment_group = node.get_comment()
        if comment_group is None:
            return

        texts: ta.List[str] = []
        target_line = check.not_none(node.get_token()).position.line
        min_comment_line = 1_000_000_000  # FIXME lol
        for comment in comment_group.comments:
            if min_comment_line > check.not_none(comment.token).position.line:
                min_comment_line = check.not_none(comment.token).position.line

            texts.append(check.not_none(comment.token).value)

        if len(texts) == 0:
            return

        comment_path = node.get_path()
        if min_comment_line < target_line:
            if isinstance(n := node, MappingYamlNode):
                if len(n.values) != 0:
                    comment_path = n.values[0].key.get_path()

            elif isinstance(n, MappingValueYamlNode):
                comment_path = n.key.get_path()

            self.add_comment_to_map(comment_path, yaml_head_comment(*texts))
        else:
            self.add_comment_to_map(comment_path, yaml_line_comment(texts[0]))

    def add_sequence_node_comment_to_map(self, node: SequenceYamlNode) -> None:
        if len(node.value_head_comments) != 0:
            for idx, hc in enumerate(node.value_head_comments):
                if hc is None:
                    continue

                texts: ta.List[str] = []
                for comment in hc.comments:
                    texts.append(check.not_none(comment.token).value)

                if len(texts) != 0:
                    self.add_comment_to_map(check.not_none(node.values[idx]).get_path(), yaml_head_comment(*texts))

        first_elem_head_comment = node.get_comment()
        if first_elem_head_comment is not None:
            texts = []
            for comment in first_elem_head_comment.comments:
                texts.append(check.not_none(comment.token).value)

            if len(texts) != 0:
                if len(node.values) != 0:
                    self.add_comment_to_map(check.not_none(node.values[0]).get_path(), yaml_head_comment(*texts))

    def add_foot_comment_to_map(self, node: YamlNode) -> None:
        fc: ta.Optional[CommentGroupYamlNode] = None
        foot_comment_path = node.get_path()

        if isinstance(n := node, SequenceYamlNode):
            fc = n.foot_comment
            if n.foot_comment is not None:
                foot_comment_path = n.foot_comment.get_path()

        elif isinstance(n, MappingYamlNode):
            fc = n.foot_comment
            if n.foot_comment is not None:
                foot_comment_path = n.foot_comment.get_path()

        elif isinstance(n, MappingValueYamlNode):
            fc = n.foot_comment
            if n.foot_comment is not None:
                foot_comment_path = n.foot_comment.get_path()

        if fc is None:
            return

        texts: ta.List[str] = []
        for comment in fc.comments:
            texts.append(check.not_none(comment.token).value)

        if len(texts) != 0:
            self.add_comment_to_map(foot_comment_path, yaml_foot_comment(*texts))

    def add_comment_to_map(self, path: str, comment: YamlComment) -> None:
        tcm = check.not_none(self.to_comment_map)[path]
        for c in tcm:
            if c.position == comment.position:
                # already added same comment
                return

        tcm.append(comment)
        tcm.sort(key=lambda c: c.position)

    def node_to_value(self, ctx: YamlDecodeContext, node: YamlNode) -> YamlErrorOr[ta.Any]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            self.set_path_comment_map(node)

            if isinstance(n := node, NullYamlNode):
                return None

            elif isinstance(n, StringYamlNode):
                return n.get_value()

            elif isinstance(n, IntegerYamlNode):
                return n.get_value()

            elif isinstance(n, FloatYamlNode):
                return n.get_value()

            elif isinstance(n, BoolYamlNode):
                return n.get_value()

            elif isinstance(n, InfinityYamlNode):
                return n.get_value()

            elif isinstance(n, NanYamlNode):
                return n.get_value()

            elif isinstance(n, TagYamlNode):
                if n.directive is not None:
                    v = self.node_to_value(ctx, check.not_none(n.value))
                    if isinstance(v, YamlError):
                        return v
                    if v is None:
                        return ''

                    return str(v)

                rtk = n.start.value
                if rtk == YamlReservedTagKeywords.TIMESTAMP:
                    t = self.cast_to_time(ctx, check.not_none(n.value))
                    if isinstance(t, YamlError):
                        return None
                    return t

                elif rtk == YamlReservedTagKeywords.INTEGER:
                    v = self.node_to_value(ctx, check.not_none(n.value))
                    if isinstance(v, YamlError):
                        return v
                    try:
                        return int(str(v))
                    except ValueError:
                        return 0

                elif rtk == YamlReservedTagKeywords.FLOAT:
                    v = self.node_to_value(ctx, check.not_none(n.value))
                    if isinstance(v, YamlError):
                        return v
                    return self.cast_to_float(v)

                elif rtk == YamlReservedTagKeywords.NULL:
                    return None

                elif rtk == YamlReservedTagKeywords.BINARY:
                    v = self.node_to_value(ctx, check.not_none(n.value))
                    if isinstance(v, YamlError):
                        return v
                    if not isinstance(v, str):
                        return YamlSyntaxError(
                            f'cannot convert {str(v)!r} to string',
                            check.not_none(check.not_none(n.value).get_token()),
                        )
                    try:
                        return base64.b64decode(v)
                    except ValueError:
                        return None

                elif rtk == YamlReservedTagKeywords.BOOLEAN:
                    v = self.node_to_value(ctx, check.not_none(n.value))
                    if isinstance(v, YamlError):
                        return v
                    l = str(v).lower()
                    if l in ('true', 't', '1', 'yes'):
                        return True
                    if l in ('false', 'f', '0', 'no'):
                        return False
                    return YamlSyntaxError(
                        f'cannot convert {v!r} to boolean',
                        check.not_none(check.not_none(n.value).get_token()),
                    )

                elif rtk == YamlReservedTagKeywords.STRING:
                    v = self.node_to_value(ctx, check.not_none(n.value))
                    if isinstance(v, YamlError):
                        return v
                    if v is None:
                        return ''
                    return str(v)

                elif rtk == YamlReservedTagKeywords.MAPPING:
                    return self.node_to_value(ctx, check.not_none(n.value))

                else:
                    return self.node_to_value(ctx, check.not_none(n.value))

            elif isinstance(n, AnchorYamlNode):
                anchor_name = check.not_none(check.not_none(n.name).get_token()).value

                # To handle the case where alias is processed recursively, the result of alias can be set to nil in
                # advance.
                self.anchor_node_map[anchor_name] = None
                anchor_value = self.node_to_value(ctx.with_anchor(anchor_name), check.not_none(n.value))
                if isinstance(anchor_value, YamlError):
                    del self.anchor_node_map[anchor_name]
                    return anchor_value
                self.anchor_node_map[anchor_name] = n.value
                self.anchor_value_map[anchor_name] = anchor_value
                return anchor_value

            elif isinstance(n, AliasYamlNode):
                text = check.not_none(n.value).string()
                if text in ctx.get_anchor_map():
                    # self recursion.
                    return None
                try:
                    v = self.anchor_value_map[text]
                except KeyError:
                    pass
                else:
                    return v
                try:
                    node2 = self.anchor_node_map[text]
                except KeyError:
                    pass
                else:
                    return self.node_to_value(ctx, check.not_none(node2))
                return YamlSyntaxError(
                    f'could not find alias {text!r}',
                    check.not_none(check.not_none(n.value).get_token()),
                )

            elif isinstance(n, LiteralYamlNode):
                return check.not_none(n.value).get_value()

            elif isinstance(n, MappingKeyYamlNode):
                return self.node_to_value(ctx, check.not_none(n.value))

            elif isinstance(n, MappingValueYamlNode):
                if n.key.is_merge_key():
                    value = self.get_map_node(check.not_none(n.value), True)
                    if isinstance(value, YamlError):
                        return value
                    it = value.map_range()
                    if self.use_ordered_map:
                        m = YamlMapSlice()
                        while it.next():
                            if (err := self.set_to_ordered_map_value(ctx, it.key_value(), m)) is not None:
                                return err
                        return m
                    m2: ta.Dict[str, ta.Any] = {}
                    while it.next():
                        if (err := self.set_to_map_value(ctx, it.key_value(), m2)) is not None:
                            return err
                    return m2

                key = self.map_key_node_to_string(ctx, n.key)
                if isinstance(key, YamlError):
                    return key

                if self.use_ordered_map:
                    v = self.node_to_value(ctx, n.value)
                    if isinstance(v, YamlError):
                        return v
                    return YamlMapSlice([YamlMapItem(key, v)])

                v = self.node_to_value(ctx, n.value)
                if isinstance(v, YamlError):
                    return v

                return {key: v}

            elif isinstance(n, MappingYamlNode):
                if self.use_ordered_map:
                    m3 = YamlMapSlice()
                    for value2 in n.values:
                        if (err := self.set_to_ordered_map_value(ctx, value2, m3)) is not None:
                            return err
                    return m3

                m4: ta.Dict[str, ta.Any] = {}
                for value3 in n.values:
                    if (err := self.set_to_map_value(ctx, value3, m4)) is not None:
                        return err
                return m4

            elif isinstance(n, SequenceYamlNode):
                v2: ta.List[ta.Any] = []
                for value4 in n.values:
                    vv = self.node_to_value(ctx, check.not_none(value4))
                    if isinstance(vv, YamlError):
                        return vv
                    v2.append(vv)
                return v2

            return None

        finally:
            self.step_out()

    def cast_to_time(self, ctx: YamlDecodeContext, src: YamlNode) -> YamlErrorOr[datetime.datetime]:
        raise NotImplementedError

    def get_map_node(self, node: YamlNode, is_merge: bool) -> YamlErrorOr[MapYamlNode]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            if isinstance(n := node, MapYamlNode):
                return n

            elif isinstance(n, AnchorYamlNode):
                anchor_name = check.not_none(check.not_none(n.name).get_token()).value
                self.anchor_node_map[anchor_name] = n.value
                return self.get_map_node(check.not_none(n.value), is_merge)

            elif isinstance(n, AliasYamlNode):
                alias_name = check.not_none(check.not_none(n.value).get_token()).value
                node2 = self.anchor_node_map[alias_name]
                if node2 is None:
                    return yaml_error(f'cannot find anchor by alias name {alias_name}')
                return self.get_map_node(node2, is_merge)

            elif isinstance(n, SequenceYamlNode):
                if not is_merge:
                    return UnexpectedNodeTypeYamlError(node.type(), YamlNodeType.MAPPING, check.not_none(node.get_token()))  # noqa
                map_nodes: ta.List[MapYamlNode] = []
                for value in n.values:
                    map_node = self.get_map_node(check.not_none(value), False)
                    if isinstance(map_node, YamlError):
                        return map_node
                    map_nodes.append(map_node)
                return yaml_sequence_merge_value(*map_nodes)

            return UnexpectedNodeTypeYamlError(node.type(), YamlNodeType.MAPPING, check.not_none(node.get_token()))

        finally:
            self.step_out()

    def get_array_node(self, node: YamlNode) -> YamlErrorOr[ta.Optional[ArrayYamlNode]]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            if isinstance(node, NullYamlNode):
                return None

            if isinstance(anchor := node, AnchorYamlNode):
                if isinstance(array_node := anchor.value, ArrayYamlNode):
                    return array_node

                return UnexpectedNodeTypeYamlError(check.not_none(anchor.value).type(), YamlNodeType.SEQUENCE, check.not_none(node.get_token()))  # noqa

            if isinstance(alias := node, AliasYamlNode):
                alias_name = check.not_none(check.not_none(alias.value).get_token()).value
                node2 = self.anchor_node_map[alias_name]
                if node2 is None:
                    return yaml_error(f'cannot find anchor by alias name {alias_name}')
                if isinstance(array_node := node2, ArrayYamlNode):
                    return array_node
                return UnexpectedNodeTypeYamlError(node2.type(), YamlNodeType.SEQUENCE, check.not_none(node2.get_token()))  # noqa

            if not isinstance(array_node := node, ArrayYamlNode):
                return UnexpectedNodeTypeYamlError(node.type(), YamlNodeType.SEQUENCE, check.not_none(node.get_token()))  # noqa

            return array_node

        finally:
            self.step_out()

    def decode_value(self, ctx: YamlDecodeContext, src: YamlNode) -> YamlErrorOr[ta.Any]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            if src.type() == YamlNodeType.ANCHOR:
                anchor = check.isinstance(src, AnchorYamlNode)
                anchor_name = check.not_none(check.not_none(anchor.name).get_token()).value
                if isinstance(av := self.decode_value(ctx.with_anchor(anchor_name), check.not_none(anchor.value)), YamlError):  # noqa
                    return av
                self.anchor_value_map[anchor_name] = av
                return None

            src_val = self.node_to_value(ctx, src)
            if isinstance(src_val, YamlError):
                return src_val

            return src_val

        finally:
            self.step_out()

    def key_to_node_map(
        self,
        ctx: YamlDecodeContext,
        node: YamlNode,
        ignore_merge_key: bool,
        get_key_or_value_node: ta.Callable[[MapYamlNodeIter], YamlNode],
    ) -> YamlErrorOr[ta.Dict[str, YamlNode]]:
        self.step_in()
        try:
            if self.is_exceeded_max_depth():
                return YamlDecodeErrors.EXCEEDED_MAX_DEPTH

            map_node = self.get_map_node(node, False)
            if isinstance(map_node, YamlError):
                return map_node
            key_map: ta.Dict[str, None] = {}
            key_to_node_map: ta.Dict[str, YamlNode] = {}
            map_iter = map_node.map_range()
            while map_iter.next():
                key_node = map_iter.key()
                if key_node.is_merge_key():
                    if ignore_merge_key:
                        continue
                    merge_map = self.key_to_node_map(ctx, map_iter.value(), ignore_merge_key, get_key_or_value_node)
                    if isinstance(merge_map, YamlError):
                        return merge_map
                    for k, v in merge_map.items():
                        if (err := self.validate_duplicate_key(key_map, k, v)) is not None:
                            return err
                        key_to_node_map[k] = v
                else:
                    key_val = self.node_to_value(ctx, key_node)
                    if isinstance(key_val, YamlError):
                        return key_val
                    if not isinstance(key := key_val, str):
                        return yaml_error('???')
                    if (err := self.validate_duplicate_key(key_map, key, key_node)) is not None:
                        return err
                    key_to_node_map[key] = get_key_or_value_node(map_iter)
            return key_to_node_map

        finally:
            self.step_out()

    def key_to_key_node_map(
        self,
        ctx: YamlDecodeContext,
        node: YamlNode,
        ignore_merge_key: bool,
    ) -> YamlErrorOr[ta.Dict[str, YamlNode]]:
        m = self.key_to_node_map(ctx, node, ignore_merge_key, lambda node_map: node_map.key())
        if isinstance(m, YamlError):
            return m
        return m

    def key_to_value_node_map(
        self,
        ctx: YamlDecodeContext,
        node: YamlNode,
        ignore_merge_key: bool,
    ) -> YamlErrorOr[ta.Dict[str, YamlNode]]:
        m = self.key_to_node_map(ctx, node, ignore_merge_key, lambda node_map: node_map.value())
        if isinstance(m, YamlError):
            return m
        return m

    # getParentMapTokenIfExists if the NodeType is a container type such as MappingType or SequenceType,
    # it is necessary to return the parent MapNode's colon token to represent the entire container.
    def get_parent_map_token_if_exists_for_validation_error(self, typ: YamlNodeType, tk: ta.Optional[YamlToken]) -> ta.Optional[YamlToken]:  # noqa
        if tk is None:
            return None
        if typ == YamlNodeType.MAPPING:
            # map:
            #   key: value
            #      ^ current token ( colon )
            if tk.prev is None:
                return tk
            key = tk.prev
            if key.prev is None:
                return tk
            return key.prev
        if typ == YamlNodeType.SEQUENCE:
            # map:
            #   - value
            #   ^ current token ( sequence entry )
            if tk.prev is None:
                return tk
            return tk.prev
        return tk

    def validate_duplicate_key(self, key_map: ta.Dict[str, None], key: ta.Any, key_node: YamlNode) -> ta.Optional[YamlError]:  # noqa
        if not isinstance(k := key, str):
            return None
        if not self.allow_duplicate_map_key:
            if k in key_map:
                return DuplicateKeyYamlError(f'duplicate key "{k}"', check.not_none(key_node.get_token()))
        key_map[k] = None
        return None

    def file_to_reader(self, file: str) -> YamlErrorOr[YamlBytesReader]:
        with open(file, 'rb') as f:
            bs = f.read()
        return ImmediateYamlBytesReader(bs)

    def is_yaml_file(self, file: str) -> bool:
        ext = file.rsplit('.', maxsplit=1)[-1]
        if ext == '.yml':
            return True
        if ext == '.yaml':
            return True
        return False

    def readers_under_dir(self, d: str) -> YamlErrorOr[ta.List[YamlBytesReader]]:
        pattern = f'{d}/*'
        matches = glob.glob(pattern)
        readers: ta.List[YamlBytesReader] = []
        for match in matches:
            if not self.is_yaml_file(match):
                continue
            if isinstance(reader := self.file_to_reader(match), YamlError):
                return reader
            readers.append(reader)
        return readers

    def readers_under_dir_recursive(self, d: str) -> YamlErrorOr[ta.List[YamlBytesReader]]:
        readers: ta.List[YamlBytesReader] = []
        for dp, _, fns in os.walk(d):
            for fn in fns:
                path = os.path.join(dp, fn)
                if not os.path.isfile(path):
                    continue
                if not self.is_yaml_file(path):
                    continue
                if isinstance(reader := self.file_to_reader(path), YamlError):
                    return reader
                readers.append(reader)
        return readers

    def resolve_reference(self, ctx: YamlDecodeContext) -> ta.Optional[YamlError]:
        for opt in self.opts:
            if (err := opt(self)) is not None:
                return err
        for file in self.reference_files:
            if isinstance(reader := self.file_to_reader(file), YamlError):
                return reader
            self.reference_readers.append(reader)
        for d in self.reference_dirs:
            if not self.is_recursive_dir:
                if isinstance(readers := self.readers_under_dir(d), YamlError):
                    return readers
                self.reference_readers.extend(readers)
            else:
                if isinstance(readers := self.readers_under_dir_recursive(d), YamlError):
                    return readers
                self.reference_readers.extend(readers)
        for reader in self.reference_readers:
            bs = reader.read()
            # assign new anchor definition to anchorMap
            if isinstance(err2 := self.parse(ctx, bs), YamlError):
                return err2
        self.is_resolved_reference = True
        return None

    def parse(self, ctx: YamlDecodeContext, bs: bytes) -> YamlErrorOr[YamlFile]:
        parse_mode: YamlParseMode = 0
        if self.to_comment_map is not None:
            parse_mode |= YAML_PARSE_COMMENTS
        opts: ta.List[YamlOption] = []
        if self.allow_duplicate_map_key:
            opts.append(yaml_allow_duplicate_map_key())
        if isinstance(f := yaml_parse_str(bs.decode(), parse_mode, *opts), YamlError):
            return f
        normalized_file = YamlFile()
        for doc in f.docs:
            # try to decode YamlNode to value and map anchor value to anchorMap
            if isinstance(v := self.node_to_value(ctx, check.not_none(doc.body)), YamlError):
                return v
            if v is not None or (doc.body is not None and doc.body.type() == YamlNodeType.NULL):
                normalized_file.docs.append(doc)
                cm = YamlCommentMap()
                cm.update(self.to_comment_map or {})
                self.comment_maps.append(cm)
            if self.to_comment_map is not None:
                self.to_comment_map.clear()
        return normalized_file

    def is_initialized(self) -> bool:
        return self.parsed_file is not None

    def decode_init(self, ctx: YamlDecodeContext) -> ta.Optional[YamlError]:
        if not self.is_resolved_reference:
            if (err := self.resolve_reference(ctx)) is not None:
                return err
        buf = self.reader.read()
        if isinstance(file := self.parse(ctx, buf), YamlError):
            return file
        self.parsed_file = file
        return None

    def _decode(self, ctx: YamlDecodeContext) -> YamlErrorOr[ta.Any]:
        self.decode_depth = 0
        self.anchor_value_map = {}
        pf = check.not_none(self.parsed_file)
        if len(pf.docs) == 0:
            # empty document.
            return None
        if len(pf.docs) <= self.stream_index:
            return EofYamlError()
        body = pf.docs[self.stream_index].body
        if body is None:
            return None
        if len(self.comment_maps) > self.stream_index:
            if (scm := self.comment_maps[self.stream_index]):
                check.not_none(self.to_comment_map).update(scm)
        if isinstance(v := self.decode_value(ctx, body), YamlError):
            return v
        self.stream_index += 1
        return v

    # Decode reads the next YAML-encoded value from its input
    # and stores it in the value pointed to by v.
    #
    # See the documentation for Unmarshal for details about the
    # conversion of YAML into a Go value.
    def decode(self) -> YamlErrorOr[ta.Any]:
        return self.decode_context(YamlDecodeContext())

    # decode_context reads the next YAML-encoded value from its input
    # and stores it in the value pointed to by v with Context.
    def decode_context(self, ctx: YamlDecodeContext) -> YamlErrorOr[ta.Any]:
        if self.is_initialized():
            if isinstance(v := self._decode(ctx), YamlError):
                return v
            return v
        if (err := self.decode_init(ctx)) is not None:
            return err
        if isinstance(v := self._decode(ctx), YamlError):
            return v
        return v

    # decode_from_node decodes node into the value pointed to by v.
    def decode_from_node(self, node: YamlNode) -> YamlErrorOr[ta.Any]:
        return self.decode_from_node_context(YamlDecodeContext(), node)

    # decode_from_node_context decodes node into the value pointed to by v with Context.
    def decode_from_node_context(self, ctx: YamlDecodeContext, node: YamlNode) -> YamlErrorOr[ta.Any]:
        if not self.is_initialized():
            if (err := self.decode_init(ctx)) is not None:
                return err
        # resolve references to the anchor on the same file
        if (err := self.node_to_value(ctx, node)) is not None:
            return err
        if isinstance(v := self.decode_value(ctx, node), YamlError):
            return v
        return v


##


def yaml_decode(s: str) -> ta.Any:
    d = YamlDecoder(ImmediateYamlBytesReader(s.encode()))
    if isinstance(v := d.decode(), YamlError):
        raise v
    return v



# @omlish-lite
from .errors import NeedMoreDataByteStreamBufferError
from .types import ByteStreamBuffer
from .types import ByteStreamBufferView


##


class ByteStreamBufferReader:
    def __init__(self, buf: ByteStreamBuffer) -> None:
        super().__init__()

        self._buf = buf

    #

    def _coalesce_exact(self, n: int) -> memoryview:
        """
        Return a contiguous view of exactly `n` readable bytes, or raise NeedMoreDataByteStreamBufferError.

        Uses `buf.coalesce(n)` to avoid per-byte Python work and to keep copying close to the buffer backend.
        """

        if n < 0:
            raise ValueError(n)
        if len(self._buf) < n:
            raise NeedMoreDataByteStreamBufferError
        mv = self._buf.coalesce(n)
        if len(mv) < n:
            # Defensive: coalesce contract should provide >= n when len(buf) >= n.
            raise NeedMoreDataByteStreamBufferError
        return mv[:n]

    #

    def peek_u8(self) -> int:
        """Peek an unsigned 8-bit integer without consuming."""

        mv = self._coalesce_exact(1)
        return mv[0]

    def read_u8(self) -> int:
        """Read and consume an unsigned 8-bit integer."""

        v = self.peek_u8()
        self._buf.advance(1)
        return v

    def peek_u16_be(self) -> int:
        """Peek an unsigned 16-bit big-endian integer without consuming."""

        mv = self._coalesce_exact(2)
        return int.from_bytes(mv, 'big', signed=False)

    def read_u16_be(self) -> int:
        """Read and consume an unsigned 16-bit big-endian integer."""

        v = self.peek_u16_be()
        self._buf.advance(2)
        return v

    def peek_u16_le(self) -> int:
        """Peek an unsigned 16-bit little-endian integer without consuming."""

        mv = self._coalesce_exact(2)
        return int.from_bytes(mv, 'little', signed=False)

    def read_u16_le(self) -> int:
        """Read and consume an unsigned 16-bit little-endian integer."""

        v = self.peek_u16_le()
        self._buf.advance(2)
        return v

    def peek_u32_be(self) -> int:
        """Peek an unsigned 32-bit big-endian integer without consuming."""

        mv = self._coalesce_exact(4)
        return int.from_bytes(mv, 'big', signed=False)

    def read_u32_be(self) -> int:
        """Read and consume an unsigned 32-bit big-endian integer."""

        v = self.peek_u32_be()
        self._buf.advance(4)
        return v

    def peek_u32_le(self) -> int:
        """Peek an unsigned 32-bit little-endian integer without consuming."""

        mv = self._coalesce_exact(4)
        return int.from_bytes(mv, 'little', signed=False)

    def read_u32_le(self) -> int:
        """Read and consume an unsigned 32-bit little-endian integer."""

        v = self.peek_u32_le()
        self._buf.advance(4)
        return v

    #

    def peek_exact(self, n: int, /) -> memoryview:
        """
        Return a contiguous view of exactly `n` readable bytes without consuming.

        Raises NeedMoreDataByteStreamBufferError if fewer than `n` bytes are currently buffered.
        """

        if n < 0:
            raise ValueError(n)
        if len(self._buf) < n:
            raise NeedMoreDataByteStreamBufferError
        mv = self._buf.coalesce(n)
        if len(mv) < n:
            raise NeedMoreDataByteStreamBufferError
        return mv[:n]

    def take(self, n: int, /) -> ByteStreamBufferView:
        """
        Consume and return a `ByteStreamBufferView`-like object representing exactly `n` bytes.

        Raises NeedMoreDataByteStreamBufferError if fewer than `n` bytes are currently buffered.
        """

        if n < 0:
            raise ValueError(n)
        if len(self._buf) < n:
            raise NeedMoreDataByteStreamBufferError
        return self._buf.split_to(n)

    def read_bytes(self, n: int, /) -> bytes:
        """
        Consume exactly `n` bytes and return them as a contiguous `bytes` object (copy boundary).

        Raises NeedMoreDataByteStreamBufferError if fewer than `n` bytes are currently buffered.
        """

        v = self.take(n)
        return v.tobytes()



import dataclasses as dc
import operator
import typing as ta


T = ta.TypeVar('T')


##


def deep_replace(o: T, *args: str | ta.Callable[[ta.Any], ta.Mapping[str, ta.Any]]) -> T:
    if not args:
        return o
    elif len(args) == 1:
        return dc.replace(o, **args[0](o))  # type: ignore
    else:
        return dc.replace(o, **{args[0]: deep_replace(getattr(o, args[0]), *args[1:])})  # type: ignore


##


def replace_if(
        fn: ta.Callable[[ta.Any, ta.Any], bool] | None,
        o: T,
        /,
        **kwargs: ta.Any,
) -> T:
    if fn is None:
        fn = lambda l, r: r is not None
    dct: dict[str, ta.Any] = {}
    for k, v in kwargs.items():
        if fn(getattr(o, k), v):
            dct[k] = v  # noqa
    if not dct:
        return o
    return dc.replace(o, **dct)  # type: ignore


def replace_ne(o: T, **kwargs: ta.Any) -> T:
    return replace_if(operator.ne, o, **kwargs)


def replace_is_not(o: T, **kwargs: ta.Any) -> T:
    return replace_if(operator.is_not, o, **kwargs)


##


def merge_if(
        fn: ta.Callable[[ta.Any, ta.Any], bool] | None,
        o: T,
        *os: T,
) -> T:
    if fn is None:
        fn = lambda l, r: r is not None
    odct = {f.name: getattr(o, f.name) for f in dc.fields(o)}  # type: ignore[arg-type]
    xdct: dict[str, ta.Any] = {}
    for x in reversed(os):
        if not odct:
            break
        if xu := {
            k: xv
            for k, ov in odct.items()
            if fn(ov, xv := getattr(x, k))
        }:
            xdct.update(xu)
            for k in xu:
                del odct[k]
    if not xdct:
        return o
    return dc.replace(o, **xdct)  # type: ignore


def merge_ne(o: T, *os: T) -> T:
    return merge_if(operator.ne, o, *os)


def merge_is_not(o: T, *os: T) -> T:
    return merge_if(operator.is_not, o, *os)



"""
TODO:
 - quote mode lol
  - general 'modes' dc
"""
import typing as ta

from ... import check
from ... import collections as col
from ... import dataclasses as dc
from ... import dispatch
from ... import lang
from ...text import parts as tp
from ..params import ParamKey
from ..params import ParamsPreparer
from ..params import ParamStyle
from ..params import make_params_preparer
from ..params import substitute_params
from .base import Node
from .binary import Binary
from .binary import BinaryOp
from .binary import BinaryOps
from .deletes import Delete
from .exprs import Literal
from .exprs import NameExpr
from .exprs import ParamExpr
from .funcs import Func
from .idents import Ident
from .inserts import Insert
from .inserts import Values
from .keywords import LiteralKeyword
from .keywords import Star
from .multi import Multi
from .multi import MultiKind
from .names import Name
from .params import Param
from .relations import Join
from .relations import JoinKind
from .relations import Table
from .selects import AllSelectItem
from .selects import ExprSelectItem
from .selects import Select
from .selects import SelectExpr
from .selects import SelectRelation
from .unary import Unary
from .unary import UnaryOp
from .unary import UnaryOps
from .unions import Union
from .updates import Field
from .updates import Fields
from .updates import Update


##


RenderedQueryParams: ta.TypeAlias = ta.Sequence[Param] | ta.Mapping[str, Param]


@dc.dataclass(frozen=True)
class RenderedQueryParts(lang.Final):
    p: tp.Part
    params: RenderedQueryParams | None
    literals: ta.Mapping[Param, ta.Any] | None = None


@dc.dataclass(frozen=True)
class RenderedQuery(lang.Final):
    s: str
    params: RenderedQueryParams | None
    literals: ta.Mapping[Param, ta.Any] | None = None


class Renderer(lang.Abstract):
    def __init__(
            self,
            *,
            param_style: ParamStyle | None = None,
            literal_style: ta.Literal['param_all', 'safe_only'] = 'safe_only',
    ) -> None:
        super().__init__()

        self._param_style = param_style
        self._literal_style = literal_style

        self._seen_params: dict[Param, int] = {}
        self._literal_params: dict[Param, ta.Any] = {}

    #

    _params_preparer_: ParamsPreparer

    def _params_preparer(self) -> ParamsPreparer:
        try:
            return self._params_preparer_
        except AttributeError:
            pass

        if (ps := self._param_style) is None:
            raise ValueError('param_style is not set')

        self._params_preparer_ = pp = make_params_preparer(ps)
        return pp

    def _see_param(self, p: Param) -> None:
        self._seen_params.setdefault(p, len(self._seen_params))

    def _param_key(self, p: Param) -> ParamKey:
        return p.n if p.n is not None else id(p)

    def rendered_params(self) -> RenderedQueryParams | None:
        if self._param_style is None:
            check.empty(self._seen_params)
            return None

        prep_params = self._params_preparer().prepare()
        subst = col.make_map_by(self._param_key, self._seen_params, strict=True)
        return substitute_params(prep_params, subst, strict=True)

    def literal_params(self) -> ta.Mapping[Param, ta.Any] | None:
        return self._literal_params or None

    def _make_literal_param(self, o: ta.Any) -> Param:
        p = Param()
        self._literal_params[p] = o
        return p

    #

    @dispatch.method(
        instance_cache=True,
        strong_dispatch_cache=True,
        uncached_dispatch_miss=True,
    )
    def render(self, o: ta.Any) -> tp.Part:
        raise TypeError(o)

    @classmethod
    def render_query_parts(cls, o: ta.Any, *args: ta.Any, **kwargs: ta.Any) -> RenderedQueryParts:
        r = cls(*args, **kwargs)
        return RenderedQueryParts(
            r.render(o),
            r.rendered_params(),
            r.literal_params(),
        )

    @classmethod
    def render_query(cls, o: ta.Any, *args: ta.Any, **kwargs: ta.Any) -> RenderedQuery:
        rqp = cls.render_query_parts(o, *args, **kwargs)
        return RenderedQuery(
            tp.render(rqp.p),
            rqp.params,
            rqp.literals,
        )


class StdRenderer(Renderer):
    # parens

    NEEDS_PAREN_TYPES: ta.AbstractSet[type[Node]] = {
        Binary,
        # IsNull,
        # SelectExpr,
    }

    def needs_paren(self, node: Node) -> bool:
        return type(node) in self.NEEDS_PAREN_TYPES

    def paren(self, node: Node) -> tp.Part:
        return tp.Wrap(self.render(node)) if self.needs_paren(node) else self.render(node)

    # binary

    BINARY_OP_TO_STR: ta.ClassVar[ta.Mapping[BinaryOp, str]] = {
        BinaryOps.EQ: '=',
        BinaryOps.NE: '!=',
        BinaryOps.LT: '<',
        BinaryOps.LE: '<=',
        BinaryOps.GT: '>',
        BinaryOps.GE: '>=',

        BinaryOps.ADD: '+',
        BinaryOps.SUB: '-',
        BinaryOps.MUL: '*',
        BinaryOps.DIV: '/',
        BinaryOps.MOD: '%',

        BinaryOps.CONCAT: '||',
    }

    @Renderer.render.register
    def render_binary(self, o: Binary) -> tp.Part:
        return [
            self.paren(o.l),
            self.BINARY_OP_TO_STR[o.op],
            self.paren(o.r),
        ]

    # deletes

    @Renderer.render.register
    def render_delete(self, o: Delete) -> tp.Part:
        return [
            'delete from',
            self.render(o.from_),
            *(['where', self.render(o.where)] if o.where is not None else []),
        ]

    # exprs

    SAFE_LITERAL_TYPES: ta.ClassVar[tuple[type, ...]] = (
        int,
    )

    @Renderer.render.register
    def render_literal(self, o: Literal) -> tp.Part:
        if self._literal_style == 'param_all':
            pass

        elif self._literal_style == 'safe_only':
            if isinstance(o.v, self.SAFE_LITERAL_TYPES):
                return str(o.v)

        else:
            raise ValueError(self._literal_style)

        return self.render(self._make_literal_param(o.v))

    @Renderer.render.register
    def render_name_expr(self, o: NameExpr) -> tp.Part:
        return self.render(o.n)

    @Renderer.render.register
    def render_param_expr(self, o: ParamExpr) -> tp.Part:
        return self.render(o.p)

    # funcs

    @Renderer.render.register
    def render_func(self, o: Func) -> tp.Part:
        args: list[tp.Part]
        if len(o.args) == 1 and isinstance(a0 := o.args[0], SelectExpr):
            args = [self.render(a0.s)]
        else:
            args = [self.render(a) for a in o.args]

        return tp.Concat([
            self.render(o.name),
            tp.Wrap(tp.List(args)),
        ])

    # idents

    @Renderer.render.register
    def render_ident(self, o: Ident) -> tp.Part:
        return f'"{o.s}"'

    # inserts

    @Renderer.render.register
    def render_values(self, o: Values) -> tp.Part:
        return [
            'values',
            tp.Wrap(tp.List([self.render(v) for v in o.vs])),
        ]

    @Renderer.render.register
    def render_insert(self, o: Insert) -> tp.Part:
        return [
            'insert into',
            self.render(o.into),
            tp.Wrap(tp.List([self.render(c) for c in o.columns])),
            self.render(o.data),
        ]

    # keywords

    @Renderer.render.register
    def render_literal_keyword(self, o: LiteralKeyword) -> tp.Part:
        return o.s

    @Renderer.render.register
    def render_star(self, o: Star) -> tp.Part:
        return '*'

    # multis

    MULTI_KIND_TO_STR: ta.ClassVar[ta.Mapping[MultiKind, str]] = {
        MultiKind.AND: 'and',
        MultiKind.OR: 'or',
    }

    @Renderer.render.register
    def render_multi(self, o: Multi) -> tp.Part:
        return tp.Wrap(tp.List(
            [self.render(e) for e in o.es],
            delimiter=' ' + self.MULTI_KIND_TO_STR[o.k],  # FIXME: Part
        ))

    # names

    @Renderer.render.register
    def render_name(self, o: Name) -> tp.Part:
        out: list[tp.Part] = []
        for n, i in enumerate(o.ps):
            if n:
                out.append('.')
            out.append(self.render(i))
        return tp.Concat(out)

    # params

    @Renderer.render.register
    def render_param(self, o: Param) -> tp.Part:
        self._see_param(o)
        return self._params_preparer().add(self._param_key(o))

    # relations

    @Renderer.render.register
    def render_table(self, o: Table) -> tp.Part:
        return [
            self.render(o.n),
            *(['as', self.render(o.a)] if o.a is not None else []),
        ]

    JOIN_KIND_TO_STR: ta.ClassVar[ta.Mapping[JoinKind, str]] = {
        JoinKind.DEFAULT: 'join',
        JoinKind.INNER: 'inner join',
        JoinKind.LEFT: 'left join',
        JoinKind.RIGHT: 'right join',
        JoinKind.FULL: 'full join',
        JoinKind.CROSS: 'cross join',
        JoinKind.NATURAL: 'natural join',
    }

    @Renderer.render.register
    def render_join(self, o: Join) -> tp.Part:
        return [
            self.render(o.l),
            self.JOIN_KIND_TO_STR[o.k],
            self.render(o.r),
            *(['on', self.render(o.c)] if o.c is not None else []),
        ]

    # selects

    @Renderer.render.register
    def render_all_select_item(self, o: AllSelectItem) -> tp.Part:
        return '*'

    @Renderer.render.register
    def render_expr_select_item(self, o: ExprSelectItem) -> tp.Part:
        return [
            self.render(o.v),
            *(['as', self.render(o.a)] if o.a is not None else []),
        ]

    @Renderer.render.register
    def render_select(self, o: Select) -> tp.Part:
        return [
            'select',
            tp.List([self.render(i) for i in o.items]),
            *(['from', self.render(o.from_)] if o.from_ is not None else []),
            *(['where', self.render(o.where)] if o.where is not None else []),
        ]

    @Renderer.render.register
    def render_select_expr(self, o: SelectExpr) -> tp.Part:
        return tp.Wrap(self.render(o.s))

    @Renderer.render.register
    def render_select_relation(self, o: SelectRelation) -> tp.Part:
        return tp.Wrap(self.render(o.s))

    # unary

    UNARY_OP_TO_STR: ta.ClassVar[ta.Mapping[UnaryOp, tuple[str, str]]] = {
        UnaryOps.NOT: ('not ', ''),
        UnaryOps.IS_NULL: ('', ' is null'),
        UnaryOps.IS_NOT_NULL: ('', ' is not null'),

        UnaryOps.POS: ('+', ''),
        UnaryOps.NEG: ('-', ''),
    }

    @Renderer.render.register
    def render_unary(self, o: Unary) -> tp.Part:
        pfx, sfx = self.UNARY_OP_TO_STR[o.op]
        return tp.Concat([
            pfx,
            self.render(o.v),
            sfx,
        ])

    # unions

    @Renderer.render.register
    def render_union(self, o: Union) -> tp.Part:
        return list(lang.interleave(
            'union all' if o.all else 'union',
            (self.render(s) for s in o.selects),
        ))

    # updates

    @Renderer.render.register
    def render_field(self, o: Field) -> tp.Part:
        return [
            self.render(o.c),
            '=',
            self.render(o.v),
        ]

    @Renderer.render.register
    def render_fields(self, o: Fields) -> tp.Part:
        return tp.List([self.render(f) for f in o.fields])

    @Renderer.render.register
    def render_update(self, o: Update) -> tp.Part:
        return [
            'update',
            self.render(o.into),
            'set',
            self.render(o.fields),
            *(['where', self.render(o.where)] if o.where is not None else []),
        ]


def render_parts(n: Node, **kwargs: ta.Any) -> RenderedQueryParts:
    return StdRenderer.render_query_parts(n, **kwargs)


def render(n: Node, **kwargs: ta.Any) -> RenderedQuery:
    return StdRenderer.render_query(n, **kwargs)



import argparse
import dataclasses as dc
import typing as ta


##


class NoExitArgumentParser(argparse.ArgumentParser):
    def __init__(
            self,
            *args: ta.Any,
            exit_on_error: bool = False,
            **kwargs: ta.Any,
    ) -> None:
        if exit_on_error:
            raise TypeError('exit_on_error=True not supported')

        super().__init__(*args, exit_on_error=False, **kwargs)  # type: ignore[misc]

    @dc.dataclass()
    class ExitNotSupportedError(BaseException):
        status: int
        message: str | None

    def exit(self, status=0, message=None):
        raise self.ExitNotSupportedError(status, message)



# ruff: noqa: UP006 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from .....io.pipelines.core import IoPipelineHandler
from .....io.pipelines.core import IoPipelineHandlerContext
from .....io.pipelines.core import IoPipelineMessages
from .....io.pipelines.flow.types import IoPipelineFlow
from .....io.pipelines.flow.types import IoPipelineFlowMessages
from .....lite.check import check
from ....headers import HttpHeaders
from ...requests import FullIoPipelineHttpRequest
from ...responses import FullIoPipelineHttpResponse
from ...responses import IoPipelineHttpResponseHead


##


@dc.dataclass(frozen=True)
class WsgiSpec:
    app: ta.Any
    host: str = '127.0.0.1'
    port: int = 8087


##


class WsgiHandler(IoPipelineHandler):
    def __init__(self, app: ta.Any) -> None:
        super().__init__()

        self._app = app

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineMessages.InitialInput):
            ctx.feed_in(msg)

            if not IoPipelineFlow.is_auto_read(ctx):
                ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())

            return

        if not isinstance(msg, FullIoPipelineHttpRequest):
            ctx.feed_in(msg)
            return

        #

        environ = {
            'REQUEST_METHOD': msg.head.method,
            'PATH_INFO': msg.head.target,
        }

        #

        started_response: ta.Optional[ta.Tuple[ta.Any, ta.Any]] = None

        def start_response(status, headers):  # noqa
            nonlocal started_response
            check.none(started_response)
            started_response = (status, headers)

        #

        ret = self._app(environ, start_response)

        #

        status, headers = check.not_none(started_response)
        status_code_str, _, status_reason = status.partition(' ')
        status_code = int(status_code_str)

        #

        body: bytes
        if isinstance(ret, bytes):
            body = ret
        elif isinstance(ret, list):
            body = b''.join(ret)
        else:
            raise TypeError(ret)

        #

        resp = FullIoPipelineHttpResponse(
            head=IoPipelineHttpResponseHead(
                status=status_code,
                reason=status_reason,
                headers=HttpHeaders(headers),
            ),
            body=body,
        )

        #

        ctx.feed_out(resp)
        ctx.feed_out(IoPipelineMessages.FinalOutput())



import typing as ta

from ... import check
from ... import dataclasses as dc
from ._compat import compat
from .cvs import Cvs
from .responses import Response
from .routes import Route
from .routes import RouteKey


if ta.TYPE_CHECKING:
    from .api import Api


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True, kw_only=True)
class AppRunParams:
    app: 'App'

    port: int | None = None


AppRunner: ta.TypeAlias = ta.Callable[[AppRunParams], None]


##


ViewFunc: ta.TypeAlias = ta.Callable[[], str]

BeforeRequestFunc: ta.TypeAlias = ta.Callable[[], Response | None]
AfterRequestFunc: ta.TypeAlias = ta.Callable[[ta.Any], ta.Any]


#


class App:
    _api: ta.ClassVar['Api']

    def __init__(
            self,
            import_name: str,
    ) -> None:
        check.not_none(self._api)

        super().__init__()

        self._import_name = check.non_empty_str(import_name)

        self._routes: set[Route] = set()
        self._routes_by_key: dict[RouteKey, Route] = {}

        self._view_funcs_by_endpoint: dict[str, ViewFunc] = {}

        self._before_request_funcs: list[BeforeRequestFunc] = []
        self._after_request_funcs: list[AfterRequestFunc] = []

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._import_name!r})'

    @property
    def import_name(self) -> str:
        return self._import_name

    ##
    # routing

    @property
    def routes(self) -> ta.AbstractSet[Route]:
        return self._routes

    @property
    def routes_by_key(self) -> ta.Mapping[RouteKey, Route]:
        return self._routes_by_key

    @property
    def view_funcs_by_endpoint(self) -> ta.Mapping[str, ViewFunc]:
        return self._view_funcs_by_endpoint

    @compat
    def get(self, rule: str, **options: ta.Any) -> ta.Callable[[T], T]:
        return self._method_route('GET', rule, **options)

    @compat
    def post(self, rule: str, **options: ta.Any) -> ta.Callable[[T], T]:
        return self._method_route('POST', rule, **options)

    @compat
    def put(self, rule: str, **options: ta.Any) -> ta.Callable[[T], T]:
        return self._method_route('PUT', rule, **options)

    @compat
    def delete(self, rule: str, **options: ta.Any) -> ta.Callable[[T], T]:
        return self._method_route('DELETE', rule, **options)

    @compat
    def patch(self, rule: str, **options: ta.Any) -> ta.Callable[[T], T]:
        return self._method_route('PATCH', rule, **options)

    def _method_route(
            self,
            method: str,
            rule: str,
            **options: ta.Any,
    ) -> ta.Callable[[T], T]:
        return self.route(
            rule,
            methods=[method],
            **options,
        )

    @compat
    def route(
            self,
            rule: str,
            **options: ta.Any,
    ) -> ta.Callable[[T], T]:
        def inner(fn):
            endpoint = options.pop('endpoint', None)
            self.add_url_rule(rule, endpoint, fn, **options)
            return fn

        return inner

    @compat
    def add_url_rule(
            self,
            rule: str,
            endpoint: str | None = None,
            view_func: ViewFunc | None = None,
            *,
            methods: ta.Iterable[str] | None = None,
    ) -> None:
        check.arg(check.non_empty_str(rule).startswith('/'))

        if endpoint is None:
            endpoint = check.not_none(view_func).__name__

        if methods is None:
            methods = ['GET']
        else:
            methods = [check.non_empty_str(m) for m in check.not_isinstance(methods, str)]
        methods = check.not_empty([s.upper() for s in methods])

        #

        route = Route(
            rule=rule,
            endpoint=endpoint,
            methods=frozenset(methods),
        )

        #

        for rk in route.keys:
            check.not_in(rk, self._routes_by_key)

        if view_func is not None:
            check.not_in(endpoint, self._view_funcs_by_endpoint)

        #

        self._routes.add(route)
        for rk in route.keys:
            self._routes_by_key[rk] = route

        if view_func is not None:
            self._view_funcs_by_endpoint[endpoint] = view_func

    ##
    # hooks

    @property
    def before_request_funcs(self) -> ta.Sequence[BeforeRequestFunc]:
        return self._before_request_funcs

    @property
    def after_request_funcs(self) -> ta.Sequence[AfterRequestFunc]:
        return self._after_request_funcs

    def before_request(self, func: T) -> T:  # BeforeRequestFunc
        self._before_request_funcs.append(ta.cast(BeforeRequestFunc, check.callable(func)))
        return func

    def after_request(self, func: T) -> T:  # AfterRequestFunc
        self._after_request_funcs.append(ta.cast(AfterRequestFunc, check.callable(func)))
        return func

    ##
    # running

    @compat
    def run(self, **kwargs: ta.Any) -> None:
        params = AppRunParams(app=self, **kwargs)
        runner = Cvs.APP_RUNNER.get()
        check.not_none(runner)(params)



# @omlish-lite
import threading
import typing as ta

from ..identities import AsyncliteIdentities


##


class SyncAsyncliteIdentities(AsyncliteIdentities):
    def current_identity(self) -> ta.Any:
        return threading.current_thread()



# @omlish-generated
# type: ignore
# ruff: noqa
# flake8: noqa
import dataclasses
import reprlib
import types


##


REGISTRY = {}


def _register(**kwargs):
    def inner(fn):
        REGISTRY[kwargs['plan_repr']] = (kwargs, fn)
        return fn
    return inner


##


@_register(
    plan_repr=(
        "Plans(tup=(CopyPlan(fields=('mode', 'adapter')), EqPlan(fields=('mode', 'adapter')), FrozenPlan(fields=('mode'"
        ", 'adapter'), allow_dynamic_dunder_attrs=False), HashPlan(action='add', fields=('mode', 'adapter'), cache=Fals"
        "e), InitPlan(fields=(InitPlan.Field(name='mode', annotation=OpRef(name='init.fields.0.annotation'), default=Op"
        "Ref(name='init.fields.0.default'), default_factory=None, init=True, override=False, field_type=FieldType.INSTA"
        "NCE, coerce=None, validate=None, check_type=None), InitPlan.Field(name='adapter', annotation=OpRef(name='init."
        "fields.1.annotation'), default=OpRef(name='init.fields.1.default'), default_factory=None, init=True, override="
        "False, field_type=FieldType.INSTANCE, coerce=None, validate=None, check_type=None)), self_param='self', std_pa"
        "rams=(), kw_only_params=('mode', 'adapter'), frozen=True, slots=False, post_init_params=None, init_fns=(), val"
        "idate_fns=()), ReprPlan(fields=(ReprPlan.Field(name='mode', kw_only=True, fn=None), ReprPlan.Field(name='adapt"
        "er', kw_only=True, fn=None)), id=False, terse=False, default_fn=None)))"
    ),
    plan_repr_sha1='5ae94c3ed2c354e407c4c2dafde59484b9d0ac88',
    op_ref_idents=(
        '__dataclass__init__fields__0__annotation',
        '__dataclass__init__fields__0__default',
        '__dataclass__init__fields__1__annotation',
        '__dataclass__init__fields__1__default',
    ),
    cls_names=(
        ('omlish.sql.api', 'AsQueryContext'),
    ),
)
def _process_dataclass__5ae94c3ed2c354e407c4c2dafde59484b9d0ac88():
    def _process_dataclass(
        *,
        __dataclass__cls,
        __dataclass__init__fields__0__annotation,
        __dataclass__init__fields__0__default,
        __dataclass__init__fields__1__annotation,
        __dataclass__init__fields__1__default,
        __dataclass__FieldFnValidationError,  # noqa
        __dataclass__FieldTypeValidationError,  # noqa
        __dataclass__FnValidationError,  # noqa
        __dataclass__FrozenInstanceError=dataclasses.FrozenInstanceError,  # noqa
        __dataclass__FunctionType=types.FunctionType,  # noqa
        __dataclass__HAS_DEFAULT_FACTORY=dataclasses._HAS_DEFAULT_FACTORY,  # noqa
        __dataclass__MISSING=dataclasses.MISSING,  # noqa
        __dataclass__None=None,  # noqa
        __dataclass__TypeError=TypeError,  # noqa
        __dataclass___recursive_repr=reprlib.recursive_repr,  # noqa
        __dataclass__isinstance=isinstance,  # noqa
        __dataclass__object_setattr=object.__setattr__,  # noqa
        __dataclass__property=property,  # noqa
    ):
        def __copy__(self):
            if self.__class__ is not __dataclass__cls:
                raise TypeError(self)
            return __dataclass__cls(  # noqa
                mode=self.mode,
                adapter=self.adapter,
            )

        __copy__.__qualname__ = f"{__dataclass__cls.__qualname__}.__copy__"
        if '__copy__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __copy__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__copy__', __copy__)

        def __eq__(self, other):
            if self is other:
                return True
            if self.__class__ is not other.__class__:
                return NotImplemented
            return (
                self.mode == other.mode and
                self.adapter == other.adapter
            )

        __eq__.__qualname__ = f"{__dataclass__cls.__qualname__}.__eq__"
        if '__eq__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __eq__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__eq__', __eq__)

        __dataclass___setattr_frozen_fields = {
            'mode',
            'adapter',
        }

        def __setattr__(self, name, value):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___setattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot assign to field {name!r}")
            super(__dataclass__cls, self).__setattr__(name, value)

        __setattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__setattr__"
        if '__setattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __setattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__setattr__', __setattr__)

        __dataclass___delattr_frozen_fields = {
            'mode',
            'adapter',
        }

        def __delattr__(self, name):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___delattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot delete field {name!r}")
            super(__dataclass__cls, self).__delattr__(name)

        __delattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__delattr__"
        if '__delattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __delattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__delattr__', __delattr__)

        def __hash__(self):
            return hash((
                self.mode,
                self.adapter,
            ))

        __hash__.__qualname__ = f"{__dataclass__cls.__qualname__}.__hash__"
        setattr(__dataclass__cls, '__hash__', __hash__)

        def __init__(
            self,
            *,
            mode: __dataclass__init__fields__0__annotation = __dataclass__init__fields__0__default,
            adapter: __dataclass__init__fields__1__annotation = __dataclass__init__fields__1__default,
        ) -> __dataclass__None:
            __dataclass__object_setattr(self, 'mode', mode)
            __dataclass__object_setattr(self, 'adapter', adapter)

        __init__.__qualname__ = f"{__dataclass__cls.__qualname__}.__init__"
        if '__init__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __init__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__init__', __init__)

        @__dataclass___recursive_repr()
        def __repr__(self):
            parts = []
            parts.append(f"mode={self.mode!r}")
            parts.append(f"adapter={self.adapter!r}")
            return (
                f"{self.__class__.__qualname__}("
                f"{', '.join(parts)}"
                f")"
            )

        __repr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__repr__"
        if '__repr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __repr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__repr__', __repr__)

    return _process_dataclass


@_register(
    plan_repr=(
        "Plans(tup=(CopyPlan(fields=('name', 'db_type', 'type')), EqPlan(fields=('name', 'db_type', 'type')), FrozenPla"
        "n(fields=('name', 'db_type', 'type'), allow_dynamic_dunder_attrs=False), HashPlan(action='add', fields=('name'"
        ", 'db_type', 'type'), cache=False), InitPlan(fields=(InitPlan.Field(name='name', annotation=OpRef(name='init.f"
        "ields.0.annotation'), default=None, default_factory=None, init=True, override=False, field_type=FieldType.INST"
        "ANCE, coerce=None, validate=None, check_type=None), InitPlan.Field(name='db_type', annotation=OpRef(name='init"
        ".fields.1.annotation'), default=OpRef(name='init.fields.1.default'), default_factory=None, init=True, override"
        "=False, field_type=FieldType.INSTANCE, coerce=None, validate=None, check_type=None), InitPlan.Field(name='type"
        "', annotation=OpRef(name='init.fields.2.annotation'), default=OpRef(name='init.fields.2.default'), default_fac"
        "tory=None, init=True, override=False, field_type=FieldType.INSTANCE, coerce=None, validate=None, check_type=No"
        "ne)), self_param='self', std_params=('name',), kw_only_params=('db_type', 'type'), frozen=True, slots=False, p"
        "ost_init_params=None, init_fns=(), validate_fns=()), ReprPlan(fields=(ReprPlan.Field(name='name', kw_only=Fals"
        "e, fn=None), ReprPlan.Field(name='db_type', kw_only=True, fn=None), ReprPlan.Field(name='type', kw_only=True, "
        "fn=None)), id=False, terse=False, default_fn=None)))"
    ),
    plan_repr_sha1='6e6d18636070554815585022cec70aa6f15ef9a9',
    op_ref_idents=(
        '__dataclass__init__fields__0__annotation',
        '__dataclass__init__fields__1__annotation',
        '__dataclass__init__fields__1__default',
        '__dataclass__init__fields__2__annotation',
        '__dataclass__init__fields__2__default',
    ),
    cls_names=(
        ('omlish.sql.api', 'Column'),
    ),
)
def _process_dataclass__6e6d18636070554815585022cec70aa6f15ef9a9():
    def _process_dataclass(
        *,
        __dataclass__cls,
        __dataclass__init__fields__0__annotation,
        __dataclass__init__fields__1__annotation,
        __dataclass__init__fields__1__default,
        __dataclass__init__fields__2__annotation,
        __dataclass__init__fields__2__default,
        __dataclass__FieldFnValidationError,  # noqa
        __dataclass__FieldTypeValidationError,  # noqa
        __dataclass__FnValidationError,  # noqa
        __dataclass__FrozenInstanceError=dataclasses.FrozenInstanceError,  # noqa
        __dataclass__FunctionType=types.FunctionType,  # noqa
        __dataclass__HAS_DEFAULT_FACTORY=dataclasses._HAS_DEFAULT_FACTORY,  # noqa
        __dataclass__MISSING=dataclasses.MISSING,  # noqa
        __dataclass__None=None,  # noqa
        __dataclass__TypeError=TypeError,  # noqa
        __dataclass___recursive_repr=reprlib.recursive_repr,  # noqa
        __dataclass__isinstance=isinstance,  # noqa
        __dataclass__object_setattr=object.__setattr__,  # noqa
        __dataclass__property=property,  # noqa
    ):
        def __copy__(self):
            if self.__class__ is not __dataclass__cls:
                raise TypeError(self)
            return __dataclass__cls(  # noqa
                name=self.name,
                db_type=self.db_type,
                type=self.type,
            )

        __copy__.__qualname__ = f"{__dataclass__cls.__qualname__}.__copy__"
        if '__copy__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __copy__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__copy__', __copy__)

        def __eq__(self, other):
            if self is other:
                return True
            if self.__class__ is not other.__class__:
                return NotImplemented
            return (
                self.name == other.name and
                self.db_type == other.db_type and
                self.type == other.type
            )

        __eq__.__qualname__ = f"{__dataclass__cls.__qualname__}.__eq__"
        if '__eq__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __eq__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__eq__', __eq__)

        __dataclass___setattr_frozen_fields = {
            'name',
            'db_type',
            'type',
        }

        def __setattr__(self, name, value):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___setattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot assign to field {name!r}")
            super(__dataclass__cls, self).__setattr__(name, value)

        __setattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__setattr__"
        if '__setattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __setattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__setattr__', __setattr__)

        __dataclass___delattr_frozen_fields = {
            'name',
            'db_type',
            'type',
        }

        def __delattr__(self, name):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___delattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot delete field {name!r}")
            super(__dataclass__cls, self).__delattr__(name)

        __delattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__delattr__"
        if '__delattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __delattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__delattr__', __delattr__)

        def __hash__(self):
            return hash((
                self.name,
                self.db_type,
                self.type,
            ))

        __hash__.__qualname__ = f"{__dataclass__cls.__qualname__}.__hash__"
        setattr(__dataclass__cls, '__hash__', __hash__)

        def __init__(
            self,
            name: __dataclass__init__fields__0__annotation,
            *,
            db_type: __dataclass__init__fields__1__annotation = __dataclass__init__fields__1__default,
            type: __dataclass__init__fields__2__annotation = __dataclass__init__fields__2__default,
        ) -> __dataclass__None:
            __dataclass__object_setattr(self, 'name', name)
            __dataclass__object_setattr(self, 'db_type', db_type)
            __dataclass__object_setattr(self, 'type', type)

        __init__.__qualname__ = f"{__dataclass__cls.__qualname__}.__init__"
        if '__init__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __init__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__init__', __init__)

        @__dataclass___recursive_repr()
        def __repr__(self):
            parts = []
            parts.append(f"name={self.name!r}")
            parts.append(f"db_type={self.db_type!r}")
            parts.append(f"type={self.type!r}")
            return (
                f"{self.__class__.__qualname__}("
                f"{', '.join(parts)}"
                f")"
            )

        __repr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__repr__"
        if '__repr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __repr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__repr__', __repr__)

    return _process_dataclass


@_register(
    plan_repr=(
        "Plans(tup=(CopyPlan(fields=('mode', 'text', 'args')), EqPlan(fields=('mode', 'text', 'args')), FrozenPlan(fiel"
        "ds=('mode', 'text', 'args'), allow_dynamic_dunder_attrs=False), HashPlan(action='add', fields=('mode', 'text',"
        " 'args'), cache=False), InitPlan(fields=(InitPlan.Field(name='mode', annotation=OpRef(name='init.fields.0.anno"
        "tation'), default=None, default_factory=None, init=True, override=False, field_type=FieldType.INSTANCE, coerce"
        "=None, validate=None, check_type=None), InitPlan.Field(name='text', annotation=OpRef(name='init.fields.1.annot"
        "ation'), default=None, default_factory=None, init=True, override=False, field_type=FieldType.INSTANCE, coerce="
        "None, validate=None, check_type=None), InitPlan.Field(name='args', annotation=OpRef(name='init.fields.2.annota"
        "tion'), default=None, default_factory=None, init=True, override=False, field_type=FieldType.INSTANCE, coerce=N"
        "one, validate=None, check_type=None)), self_param='self', std_params=('mode', 'text', 'args'), kw_only_params="
        "(), frozen=True, slots=False, post_init_params=None, init_fns=(), validate_fns=()), ReprPlan(fields=(ReprPlan."
        "Field(name='mode', kw_only=False, fn=None), ReprPlan.Field(name='text', kw_only=False, fn=None), ReprPlan.Fiel"
        "d(name='args', kw_only=False, fn=None)), id=False, terse=False, default_fn=None)))"
    ),
    plan_repr_sha1='0f274d2cea40d870bb1ddb08940cd936fbf23d04',
    op_ref_idents=(
        '__dataclass__init__fields__0__annotation',
        '__dataclass__init__fields__1__annotation',
        '__dataclass__init__fields__2__annotation',
    ),
    cls_names=(
        ('omlish.sql.api', 'Query'),
    ),
)
def _process_dataclass__0f274d2cea40d870bb1ddb08940cd936fbf23d04():
    def _process_dataclass(
        *,
        __dataclass__cls,
        __dataclass__init__fields__0__annotation,
        __dataclass__init__fields__1__annotation,
        __dataclass__init__fields__2__annotation,
        __dataclass__FieldFnValidationError,  # noqa
        __dataclass__FieldTypeValidationError,  # noqa
        __dataclass__FnValidationError,  # noqa
        __dataclass__FrozenInstanceError=dataclasses.FrozenInstanceError,  # noqa
        __dataclass__FunctionType=types.FunctionType,  # noqa
        __dataclass__HAS_DEFAULT_FACTORY=dataclasses._HAS_DEFAULT_FACTORY,  # noqa
        __dataclass__MISSING=dataclasses.MISSING,  # noqa
        __dataclass__None=None,  # noqa
        __dataclass__TypeError=TypeError,  # noqa
        __dataclass___recursive_repr=reprlib.recursive_repr,  # noqa
        __dataclass__isinstance=isinstance,  # noqa
        __dataclass__object_setattr=object.__setattr__,  # noqa
        __dataclass__property=property,  # noqa
    ):
        def __copy__(self):
            if self.__class__ is not __dataclass__cls:
                raise TypeError(self)
            return __dataclass__cls(  # noqa
                mode=self.mode,
                text=self.text,
                args=self.args,
            )

        __copy__.__qualname__ = f"{__dataclass__cls.__qualname__}.__copy__"
        if '__copy__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __copy__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__copy__', __copy__)

        def __eq__(self, other):
            if self is other:
                return True
            if self.__class__ is not other.__class__:
                return NotImplemented
            return (
                self.mode == other.mode and
                self.text == other.text and
                self.args == other.args
            )

        __eq__.__qualname__ = f"{__dataclass__cls.__qualname__}.__eq__"
        if '__eq__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __eq__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__eq__', __eq__)

        __dataclass___setattr_frozen_fields = {
            'mode',
            'text',
            'args',
        }

        def __setattr__(self, name, value):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___setattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot assign to field {name!r}")
            super(__dataclass__cls, self).__setattr__(name, value)

        __setattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__setattr__"
        if '__setattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __setattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__setattr__', __setattr__)

        __dataclass___delattr_frozen_fields = {
            'mode',
            'text',
            'args',
        }

        def __delattr__(self, name):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___delattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot delete field {name!r}")
            super(__dataclass__cls, self).__delattr__(name)

        __delattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__delattr__"
        if '__delattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __delattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__delattr__', __delattr__)

        def __hash__(self):
            return hash((
                self.mode,
                self.text,
                self.args,
            ))

        __hash__.__qualname__ = f"{__dataclass__cls.__qualname__}.__hash__"
        setattr(__dataclass__cls, '__hash__', __hash__)

        def __init__(
            self,
            mode: __dataclass__init__fields__0__annotation,
            text: __dataclass__init__fields__1__annotation,
            args: __dataclass__init__fields__2__annotation,
        ) -> __dataclass__None:
            __dataclass__object_setattr(self, 'mode', mode)
            __dataclass__object_setattr(self, 'text', text)
            __dataclass__object_setattr(self, 'args', args)

        __init__.__qualname__ = f"{__dataclass__cls.__qualname__}.__init__"
        if '__init__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __init__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__init__', __init__)

        @__dataclass___recursive_repr()
        def __repr__(self):
            parts = []
            parts.append(f"mode={self.mode!r}")
            parts.append(f"text={self.text!r}")
            parts.append(f"args={self.args!r}")
            return (
                f"{self.__class__.__qualname__}("
                f"{', '.join(parts)}"
                f")"
            )

        __repr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__repr__"
        if '__repr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __repr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__repr__', __repr__)

    return _process_dataclass


@_register(
    plan_repr=(
        "Plans(tup=(CopyPlan(fields=('columns', 'values')), EqPlan(fields=('columns', 'values')), FrozenPlan(fields=('c"
        "olumns', 'values'), allow_dynamic_dunder_attrs=False), HashPlan(action='add', fields=('columns', 'values'), ca"
        "che=False), InitPlan(fields=(InitPlan.Field(name='columns', annotation=OpRef(name='init.fields.0.annotation'),"
        " default=None, default_factory=None, init=True, override=False, field_type=FieldType.INSTANCE, coerce=None, va"
        "lidate=None, check_type=None), InitPlan.Field(name='values', annotation=OpRef(name='init.fields.1.annotation')"
        ", default=None, default_factory=None, init=True, override=False, field_type=FieldType.INSTANCE, coerce=None, v"
        "alidate=None, check_type=None)), self_param='self', std_params=('columns', 'values'), kw_only_params=(), froze"
        "n=True, slots=False, post_init_params=(), init_fns=(), validate_fns=()), ReprPlan(fields=(ReprPlan.Field(name="
        "'columns', kw_only=False, fn=None), ReprPlan.Field(name='values', kw_only=False, fn=None)), id=False, terse=Fa"
        "lse, default_fn=None)))"
    ),
    plan_repr_sha1='afba95cc077e24ab1876f8a8bb8cea66aee709a9',
    op_ref_idents=(
        '__dataclass__init__fields__0__annotation',
        '__dataclass__init__fields__1__annotation',
    ),
    cls_names=(
        ('omlish.sql.api', 'Row'),
    ),
)
def _process_dataclass__afba95cc077e24ab1876f8a8bb8cea66aee709a9():
    def _process_dataclass(
        *,
        __dataclass__cls,
        __dataclass__init__fields__0__annotation,
        __dataclass__init__fields__1__annotation,
        __dataclass__FieldFnValidationError,  # noqa
        __dataclass__FieldTypeValidationError,  # noqa
        __dataclass__FnValidationError,  # noqa
        __dataclass__FrozenInstanceError=dataclasses.FrozenInstanceError,  # noqa
        __dataclass__FunctionType=types.FunctionType,  # noqa
        __dataclass__HAS_DEFAULT_FACTORY=dataclasses._HAS_DEFAULT_FACTORY,  # noqa
        __dataclass__MISSING=dataclasses.MISSING,  # noqa
        __dataclass__None=None,  # noqa
        __dataclass__TypeError=TypeError,  # noqa
        __dataclass___recursive_repr=reprlib.recursive_repr,  # noqa
        __dataclass__isinstance=isinstance,  # noqa
        __dataclass__object_setattr=object.__setattr__,  # noqa
        __dataclass__property=property,  # noqa
    ):
        def __copy__(self):
            if self.__class__ is not __dataclass__cls:
                raise TypeError(self)
            return __dataclass__cls(  # noqa
                columns=self.columns,
                values=self.values,
            )

        __copy__.__qualname__ = f"{__dataclass__cls.__qualname__}.__copy__"
        if '__copy__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __copy__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__copy__', __copy__)

        def __eq__(self, other):
            if self is other:
                return True
            if self.__class__ is not other.__class__:
                return NotImplemented
            return (
                self.columns == other.columns and
                self.values == other.values
            )

        __eq__.__qualname__ = f"{__dataclass__cls.__qualname__}.__eq__"
        if '__eq__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __eq__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__eq__', __eq__)

        __dataclass___setattr_frozen_fields = {
            'columns',
            'values',
        }

        def __setattr__(self, name, value):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___setattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot assign to field {name!r}")
            super(__dataclass__cls, self).__setattr__(name, value)

        __setattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__setattr__"
        if '__setattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __setattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__setattr__', __setattr__)

        __dataclass___delattr_frozen_fields = {
            'columns',
            'values',
        }

        def __delattr__(self, name):
            if (
                type(self) is __dataclass__cls
                or name in __dataclass___delattr_frozen_fields
            ):
                raise __dataclass__FrozenInstanceError(f"cannot delete field {name!r}")
            super(__dataclass__cls, self).__delattr__(name)

        __delattr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__delattr__"
        if '__delattr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __delattr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__delattr__', __delattr__)

        def __hash__(self):
            return hash((
                self.columns,
                self.values,
            ))

        __hash__.__qualname__ = f"{__dataclass__cls.__qualname__}.__hash__"
        setattr(__dataclass__cls, '__hash__', __hash__)

        def __init__(
            self,
            columns: __dataclass__init__fields__0__annotation,
            values: __dataclass__init__fields__1__annotation,
        ) -> __dataclass__None:
            __dataclass__object_setattr(self, 'columns', columns)
            __dataclass__object_setattr(self, 'values', values)
            self.__post_init__()

        __init__.__qualname__ = f"{__dataclass__cls.__qualname__}.__init__"
        if '__init__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __init__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__init__', __init__)

        @__dataclass___recursive_repr()
        def __repr__(self):
            parts = []
            parts.append(f"columns={self.columns!r}")
            parts.append(f"values={self.values!r}")
            return (
                f"{self.__class__.__qualname__}("
                f"{', '.join(parts)}"
                f")"
            )

        __repr__.__qualname__ = f"{__dataclass__cls.__qualname__}.__repr__"
        if '__repr__' in __dataclass__cls.__dict__:
            raise __dataclass__TypeError(f"Cannot overwrite attribute __repr__ in class {__dataclass__cls.__name__}")
        setattr(__dataclass__cls, '__repr__', __repr__)

    return _process_dataclass



import typing as ta

from .. import check
from .. import lang
from .. import reflect as rfl
from .values import TypedValue


##


def reflect_typed_values_impls(
        *rtys: rfl.Type,
        find_abstract_subclasses: bool = False,
) -> set[type[TypedValue]]:
    tv_cls_set: set[type[TypedValue]] = set()

    todo = list(rtys)
    seen = set()
    while todo:
        cur = todo.pop()
        if cur in seen:
            continue
        seen.add(cur)

        if isinstance(cur, rfl.Union):
            todo.extend(cur.args)

        elif isinstance(cur, ta.TypeVar):
            todo.append(rfl.type_(rfl.get_type_var_bound(cur)))

        elif isinstance(cur, type):
            cur = check.issubclass(check.isinstance(cur, type), TypedValue)

            if find_abstract_subclasses and lang.is_abstract_class(cur):
                todo.extend(lang.deep_subclasses(cur, concrete_only=True))

            else:
                tv_cls_set.add(cur)

        else:
            raise TypeError(cur)

    return tv_cls_set



# ruff: noqa: UP006
# @omlish-lite
import logging
import typing as ta

from .asyncs import LoggerToAsyncLogger
from .base import AsyncLogger
from .base import Logger
from .std.loggers import StdLogger


##


def _get_module_std_logger(mod_globals: ta.Mapping[str, ta.Any]) -> logging.Logger:
    return logging.getLogger(mod_globals.get('__name__'))


def get_module_logger(mod_globals: ta.Mapping[str, ta.Any]) -> Logger:
    return StdLogger(_get_module_std_logger(mod_globals))


def get_module_async_logger(mod_globals: ta.Mapping[str, ta.Any]) -> AsyncLogger:
    return LoggerToAsyncLogger(get_module_logger(mod_globals))


def get_module_loggers(mod_globals: ta.Mapping[str, ta.Any]) -> ta.Tuple[Logger, AsyncLogger]:
    return (
        log := get_module_logger(mod_globals),
        LoggerToAsyncLogger(log),
    )



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from .base import CompositeOp
from .base import LeafOp
from .base import Op
from .base import OpTuple


##


class Literal(LeafOp, lang.Abstract):
    def _match_repr(self) -> str:
        return repr(self)


@ta.final
class StringLiteral(Literal, lang.Final):
    def __init__(self, value: str) -> None:
        super().__init__()

        self._value = check.non_empty_str(value)

    @property
    def value(self) -> str:
        return self._value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._value!r})'


@ta.final
class CaseInsensitiveStringLiteral(Literal, lang.Final):
    def __init__(self, value: str) -> None:
        super().__init__()

        self._value = check.non_empty_str(value).casefold()

    @property
    def value(self) -> str:
        return self._value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._value!r})'


@ta.final
class RangeLiteral(Literal, lang.Final):
    @dc.dataclass(frozen=True)
    class Range:
        lo: str
        hi: str

        def __post_init__(self) -> None:
            for c in (self.lo, self.hi):
                check.equal(len(check.non_empty_str(c)), 1)
            check.state(self.hi >= self.lo)

    def __init__(self, value: Range) -> None:
        super().__init__()

        self._value = check.isinstance(value, RangeLiteral.Range)

    @property
    def value(self) -> Range:
        return self._value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._value!r})'


@ta.overload
def literal(s: str, *, case_sensitive: bool = False) -> StringLiteral:
    ...


@ta.overload
def literal(lo: str, hi: str) -> RangeLiteral:
    ...


def literal(*args, case_sensitive=None):
    if not args:
        raise TypeError
    elif len(args) == 1:
        s = check.isinstance(check.single(args), str)
        if case_sensitive:
            return StringLiteral(s)
        else:
            return CaseInsensitiveStringLiteral(s)
    elif len(args) == 2:
        check.none(case_sensitive)
        return RangeLiteral(RangeLiteral.Range(*map(check.of_isinstance(str), args)))
    else:
        raise TypeError(args)


##


@ta.final
class Concat(CompositeOp, lang.Final):
    def __init__(self, *children: Op) -> None:
        super().__init__()

        check.arg(len(children) > 1)
        for i, c in enumerate(children):
            check.isinstance(c, Op)
            if i:
                check.state(not (isinstance(c, Concat) and isinstance(children[i - 1], Concat)))
        self._children = children

    @property
    def children(self) -> OpTuple:
        return self._children

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({", ".join(map(repr, self._children))})'

    def replace_children(self, *children: Op) -> Op:
        if children == self._children:
            return self

        return concat(*children)


def concat(*children: Op) -> Op:
    if len(children) == 1:
        return children[0]

    check.not_empty(children)

    lst: list[Op | list[Op]] = []
    for c in children:
        if (
                lst and
                isinstance(c, Concat) and
                isinstance(ll := lst[-1], (Concat, list))
        ):
            if isinstance(ll, list):
                ll.extend(c.children)
            else:
                lst.append([*ta.cast(list, lst.pop()), *c.children])
        else:
            lst.append(c)

    if len(lst) == 1:
        return Concat(*e) if isinstance(e := lst[0], list) else e

    return Concat(*[Concat(*e) if isinstance(e, list) else e for e in lst])


##


@ta.final
class Repeat(CompositeOp, lang.Final):
    @dc.dataclass(frozen=True)
    class Times:
        min: int = 0
        max: int | None = None

        def __post_init__(self) -> None:
            if self.max is not None:
                check.state(self.max >= self.min)

        def __repr__(self) -> str:
            if not self.min and self.max is None:
                s = '*'
            elif not self.min and self.max == 1:
                s = '?'
            elif self.min and self.max is None:
                s = f'{self.min}+'
            else:
                s = f'{self.min}-{self.max!r}'
            return f'{self.__class__.__name__}({s})'

    def __init__(self, times: Times, child: Op) -> None:
        super().__init__()

        self._times = check.isinstance(times, Repeat.Times)
        self._child = check.isinstance(child, Op)

    @property
    def times(self) -> Times:
        return self._times

    @property
    def child(self) -> Op:
        return self._child

    @property
    def children(self) -> OpTuple:
        return (self._child,)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._times}, {self._child!r})'

    def replace_children(self, *children: Op) -> Op:
        child = check.single(children)
        if child == self._child:
            return self

        return Repeat(self._times, child)


@ta.overload
def repeat(child: Op) -> Repeat:  # noqa
    ...


@ta.overload
def repeat(times: Repeat.Times, child: Op) -> Repeat:  # noqa
    ...


@ta.overload
def repeat(min: int, child: Op) -> Repeat:  # noqa
    ...


@ta.overload
def repeat(min: int, max: int | None, child: Op) -> Repeat:  # noqa
    ...


def repeat(*args):
    min: int  # noqa
    max: int | None  # noqa

    if len(args) < 2:
        [child] = args
        min, max = 0, None  # noqa

    elif len(args) > 2:
        min, max, child = args  # noqa

    else:
        ti, child = args  # noqa

        if isinstance(ti, Repeat.Times):
            min, max = ti.min, ti.max  # noqa

        else:
            min, max = ti, None  # noqa

    return Repeat(
        Repeat.Times(
            check.isinstance(min, int),
            check.isinstance(max, (int, None)),
        ),
        check.isinstance(child, Op),
    )


ZERO_OR_ONE_TIMES = Repeat.Times(0, 1)


def option(child: Op) -> Repeat:
    return Repeat(ZERO_OR_ONE_TIMES, check.isinstance(child, Op))


##


@ta.final
class Either(CompositeOp, lang.Final):
    def __init__(self, *children: Op, first_match: bool = False) -> None:
        super().__init__()

        for c in check.not_empty(children):
            check.isinstance(c, Op)
        self._children = children
        self._first_match = first_match

    @property
    def children(self) -> OpTuple:
        return self._children

    @property
    def first_match(self) -> bool:
        return self._first_match

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}@{id(self):x}('
            f'{", ".join(map(repr, self._children))}'
            f'{", first_match=True" if self._first_match else ""})'
        )

    def replace_children(self, *children: Op) -> Op:
        if children == self._children:
            return self

        return Either(*children, first_match=self._first_match)


either = Either


##


@ta.final
class RuleRef(Op, lang.Final):
    def __init__(self, name: str) -> None:
        super().__init__()

        self._name = check.non_empty_str(name)

        self._name_f = name.casefold()

    def coalesce(self, other: Op) -> Op:
        """
        Op nodes are compared by identity, and transformations return identical node instances when nothing has changed.
        This method assists with that, preserving RuleRef node identity if the given node is otherwise equal.
        """

        if isinstance(other, RuleRef) and other.name_f == self.name_f:
            return self

        return other

    @property
    def name(self) -> str:
        return self._name

    @property
    def name_f(self) -> str:
        return self._name_f

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._name!r})'

    def _match_repr(self) -> str:
        return repr(self)


rule = RuleRef



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import contextlib
import socket
import typing as ta

from ..lite.check import check
from ..lite.timeouts import Timeout
from ..lite.timeouts import TimeoutLike


##


DEFAULT_AVAILABLE_PORT_HOST: str = '127.0.0.1'


@contextlib.contextmanager
def get_available_port_context(
        host: ta.Optional[str] = None,
        family: int = socket.AF_INET,
        type: int = socket.SOCK_STREAM,  # noqa
        *,
        listen: ta.Union[bool, int, None] = False,
) -> ta.Iterator[int]:
    if host is None:
        host = DEFAULT_AVAILABLE_PORT_HOST

    if listen is not None:
        if listen is True:
            listen = 1
        elif listen is False:
            listen = None
        else:
            listen = check.isinstance(listen, int)

    with socket.socket(family, type) as sock:
        sock.bind((host, 0))

        if listen is not None:
            sock.listen(listen)

        port = sock.getsockname()[1]

        yield port


def get_available_port(*args: ta.Any, **kwargs: ta.Any) -> int:
    with get_available_port_context(*args, **kwargs) as port:
        return port


def get_available_ports(
        n: int,
        *,
        host: ta.Optional[str] = None,
        exclude: ta.Optional[ta.Iterable[int]] = None,
        timeout: TimeoutLike = None,
) -> ta.List[int]:
    exclude = set(exclude or [])

    seen: ta.Set[int] = set()
    ret: ta.List[int] = []

    timeout = Timeout.of(timeout)

    with contextlib.ExitStack() as es:
        while len(ret) < n:
            timeout()

            cur = es.enter_context(get_available_port_context(host))

            check.not_in(cur, seen)
            seen.add(cur)

            if cur not in exclude:
                ret.append(cur)

    return ret



# @omlish-lite
from ....lite.abstract import Abstract
from ..base import AsyncliteApi
from ..base import AsyncliteObject


##


class SyncAsyncliteBase(Abstract):
    pass


class SyncAsyncliteObject(AsyncliteObject, SyncAsyncliteBase, Abstract):
    pass


class SyncAsyncliteApi(AsyncliteApi, SyncAsyncliteBase, Abstract):
    pass



# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
# https://github.com/python/cpython/blob/ce4b0ede16aea62ee7b1e02df7e1538102a356da/Lib/_pyrepl/terminfo.py
"""Pure Python curses-like terminal capability queries."""
import dataclasses as dc
import errno
import os
import pathlib
import re
import struct
import typing as ta


##


# Special values for absent/cancelled capabilities
ABSENT_BOOLEAN = -1
ABSENT_NUMERIC = -1
CANCELLED_NUMERIC = -2
ABSENT_STRING = None
CANCELLED_STRING = None


##


# Hard-coded terminal capabilities for common terminals. This is a minimal subset needed by PyREPL.
_FALLBACK_TERMINAL_CAPABILITIES: ta.Mapping[str, ta.Mapping[str, bytes]] = {

    # ANSI/xterm-compatible terminals
    'ansi': {
        # Bell
        'bel': b'\x07',

        # Cursor movement
        'cub': b'\x1b[%p1%dD',          # Move cursor left N columns
        'cud': b'\x1b[%p1%dB',          # Move cursor down N rows
        'cuf': b'\x1b[%p1%dC',          # Move cursor right N columns
        'cuu': b'\x1b[%p1%dA',          # Move cursor up N rows
        'cub1': b'\x08',                # Move cursor left 1 column
        'cud1': b'\n',                  # Move cursor down 1 row
        'cuf1': b'\x1b[C',              # Move cursor right 1 column
        'cuu1': b'\x1b[A',              # Move cursor up 1 row
        'cup': b'\x1b[%i%p1%d;%p2%dH',  # Move cursor to row, column
        'hpa': b'\x1b[%i%p1%dG',        # Move cursor to column
        # Clear operations
        'clear': b'\x1b[H\x1b[2J',  # Clear screen and home cursor
        'el': b'\x1b[K',            # Clear to end of line

        # Insert/delete
        'dch': b'\x1b[%p1%dP',  # Delete N characters
        'dch1': b'\x1b[P',      # Delete 1 character
        'ich': b'\x1b[%p1%d@',  # Insert N characters
        'ich1': b'',            # Insert 1 character

        # Cursor visibility
        'civis': b'\x1b[?25l',           # Make cursor invisible
        'cnorm': b'\x1b[?12l\x1b[?25h',  # Make cursor normal (visible)

        # Scrolling
        'ind': b'\n',    # Scroll up one line
        'ri': b'\x1bM',  # Scroll down one line

        # Keypad mode
        'smkx': b'\x1b[?1h\x1b=',  # Enable keypad mode
        'rmkx': b'\x1b[?1l\x1b>',  # Disable keypad mode

        # Padding (not used in modern terminals)
        'pad': b'',

        # Function keys and special keys
        'kdch1': b'\x1b[3~',  # Delete key
        'kcud1': b'\x1bOB',   # Down arrow
        'kend': b'\x1bOF',    # End key
        'kent': b'\x1bOM',    # Enter key
        'khome': b'\x1bOH',   # Home key
        'kich1': b'\x1b[2~',  # Insert key
        'kcub1': b'\x1bOD',   # Left arrow
        'knp': b'\x1b[6~',    # Page down
        'kpp': b'\x1b[5~',    # Page up
        'kcuf1': b'\x1bOC',   # Right arrow
        'kcuu1': b'\x1bOA',   # Up arrow

        # Function keys F1-F20
        'kf1': b'\x1bOP',
        'kf2': b'\x1bOQ',
        'kf3': b'\x1bOR',
        'kf4': b'\x1bOS',
        'kf5': b'\x1b[15~',
        'kf6': b'\x1b[17~',
        'kf7': b'\x1b[18~',
        'kf8': b'\x1b[19~',
        'kf9': b'\x1b[20~',
        'kf10': b'\x1b[21~',
        'kf11': b'\x1b[23~',
        'kf12': b'\x1b[24~',
        'kf13': b'\x1b[1;2P',
        'kf14': b'\x1b[1;2Q',
        'kf15': b'\x1b[1;2R',
        'kf16': b'\x1b[1;2S',
        'kf17': b'\x1b[15;2~',
        'kf18': b'\x1b[17;2~',
        'kf19': b'\x1b[18;2~',
        'kf20': b'\x1b[19;2~',
    },

    # Dumb terminal - minimal capabilities
    'dumb': {
        'bel': b'\x07',  # Bell
        'cud1': b'\n',   # Move down 1 row (newline)
        'ind': b'\n',    # Scroll up one line (newline)
    },

    # Linux console
    'linux': {
        # Bell
        'bel': b'\x07',

        # Cursor movement
        'cub': b'\x1b[%p1%dD',          # Move cursor left N columns
        'cud': b'\x1b[%p1%dB',          # Move cursor down N rows
        'cuf': b'\x1b[%p1%dC',          # Move cursor right N columns
        'cuu': b'\x1b[%p1%dA',          # Move cursor up N rows
        'cub1': b'\x08',                # Move cursor left 1 column (backspace)
        'cud1': b'\n',                  # Move cursor down 1 row (newline)
        'cuf1': b'\x1b[C',              # Move cursor right 1 column
        'cuu1': b'\x1b[A',              # Move cursor up 1 row
        'cup': b'\x1b[%i%p1%d;%p2%dH',  # Move cursor to row, column
        'hpa': b'\x1b[%i%p1%dG',        # Move cursor to column

        # Clear operations
        'clear': b'\x1b[H\x1b[J',  # Clear screen and home cursor (different from ansi!)
        'el': b'\x1b[K',           # Clear to end of line

        # Insert/delete
        'dch': b'\x1b[%p1%dP',  # Delete N characters
        'dch1': b'\x1b[P',      # Delete 1 character
        'ich': b'\x1b[%p1%d@',  # Insert N characters
        'ich1': b'\x1b[@',      # Insert 1 character

        # Cursor visibility
        'civis': b'\x1b[?25l\x1b[?1c',  # Make cursor invisible
        'cnorm': b'\x1b[?25h\x1b[?0c',  # Make cursor normal

        # Scrolling
        'ind': b'\n',    # Scroll up one line
        'ri': b'\x1bM',  # Scroll down one line

        # Keypad mode
        'smkx': b'\x1b[?1h\x1b=',  # Enable keypad mode
        'rmkx': b'\x1b[?1l\x1b>',  # Disable keypad mode

        # Function keys and special keys
        'kdch1': b'\x1b[3~',  # Delete key
        'kcud1': b'\x1b[B',   # Down arrow
        'kend': b'\x1b[4~',   # End key (different from ansi!)
        'khome': b'\x1b[1~',  # Home key (different from ansi!)
        'kich1': b'\x1b[2~',  # Insert key
        'kcub1': b'\x1b[D',   # Left arrow
        'knp': b'\x1b[6~',    # Page down
        'kpp': b'\x1b[5~',    # Page up
        'kcuf1': b'\x1b[C',   # Right arrow
        'kcuu1': b'\x1b[A',   # Up arrow

        # Function keys
        'kf1': b'\x1b[[A',
        'kf2': b'\x1b[[B',
        'kf3': b'\x1b[[C',
        'kf4': b'\x1b[[D',
        'kf5': b'\x1b[[E',
        'kf6': b'\x1b[17~',
        'kf7': b'\x1b[18~',
        'kf8': b'\x1b[19~',
        'kf9': b'\x1b[20~',
        'kf10': b'\x1b[21~',
        'kf11': b'\x1b[23~',
        'kf12': b'\x1b[24~',
        'kf13': b'\x1b[25~',
        'kf14': b'\x1b[26~',
        'kf15': b'\x1b[28~',
        'kf16': b'\x1b[29~',
        'kf17': b'\x1b[31~',
        'kf18': b'\x1b[32~',
        'kf19': b'\x1b[33~',
        'kf20': b'\x1b[34~',
    },

}


# Map common TERM values to capability sets
_TERM_ALIASES: ta.Mapping[str, str] = {
    'xterm': 'ansi',
    'xterm-color': 'ansi',
    'xterm-256color': 'ansi',
    'screen': 'ansi',
    'screen-256color': 'ansi',
    'tmux': 'ansi',
    'tmux-256color': 'ansi',
    'vt100': 'ansi',
    'vt220': 'ansi',
    'rxvt': 'ansi',
    'rxvt-unicode': 'ansi',
    'rxvt-unicode-256color': 'ansi',
    'unknown': 'dumb',
}


# Terminfo constants
_MAGIC16 = 0o432   # Magic number for 16-bit terminfo format
_MAGIC32 = 0o1036  # Magic number for 32-bit terminfo format


# Standard string capability names from ncurses Caps file This matches the order used by ncurses when compiling terminfo
_ORDERED_STRING_NAMES: tuple[str, ...] = (
    'cbt',
    'bel',
    'cr',
    'csr',
    'tbc',
    'clear',
    'el',
    'ed',
    'hpa',
    'cmdch',
    'cup',
    'cud1',
    'home',
    'civis',
    'cub1',
    'mrcup',
    'cnorm',
    'cuf1',
    'll',
    'cuu1',
    'cvvis',
    'dch1',
    'dl1',
    'dsl',
    'hd',
    'smacs',
    'blink',
    'bold',
    'smcup',
    'smdc',
    'dim',
    'smir',
    'invis',
    'prot',
    'rev',
    'smso',
    'smul',
    'ech',
    'rmacs',
    'sgr0',
    'rmcup',
    'rmdc',
    'rmir',
    'rmso',
    'rmul',
    'flash',
    'ff',
    'fsl',
    'is1',
    'is2',
    'is3',
    'if',
    'ich1',
    'il1',
    'ip',
    'kbs',
    'ktbc',
    'kclr',
    'kctab',
    'kdch1',
    'kdl1',
    'kcud1',
    'krmir',
    'kel',
    'ked',
    'kf0',
    'kf1',
    'kf10',
    'kf2',
    'kf3',
    'kf4',
    'kf5',
    'kf6',
    'kf7',
    'kf8',
    'kf9',
    'khome',
    'kich1',
    'kil1',
    'kcub1',
    'kll',
    'knp',
    'kpp',
    'kcuf1',
    'kind',
    'kri',
    'khts',
    'kcuu1',
    'rmkx',
    'smkx',
    'lf0',
    'lf1',
    'lf10',
    'lf2',
    'lf3',
    'lf4',
    'lf5',
    'lf6',
    'lf7',
    'lf8',
    'lf9',
    'rmm',
    'smm',
    'nel',
    'pad',
    'dch',
    'dl',
    'cud',
    'ich',
    'indn',
    'il',
    'cub',
    'cuf',
    'rin',
    'cuu',
    'pfkey',
    'pfloc',
    'pfx',
    'mc0',
    'mc4',
    'mc5',
    'rep',
    'rs1',
    'rs2',
    'rs3',
    'rf',
    'rc',
    'vpa',
    'sc',
    'ind',
    'ri',
    'sgr',
    'hts',
    'wind',
    'ht',
    'tsl',
    'uc',
    'hu',
    'iprog',
    'ka1',
    'ka3',
    'kb2',
    'kc1',
    'kc3',
    'mc5p',
    'rmp',
    'acsc',
    'pln',
    'kcbt',
    'smxon',
    'rmxon',
    'smam',
    'rmam',
    'xonc',
    'xoffc',
    'enacs',
    'smln',
    'rmln',
    'kbeg',
    'kcan',
    'kclo',
    'kcmd',
    'kcpy',
    'kcrt',
    'kend',
    'kent',
    'kext',
    'kfnd',
    'khlp',
    'kmrk',
    'kmsg',
    'kmov',
    'knxt',
    'kopn',
    'kopt',
    'kprv',
    'kprt',
    'krdo',
    'kref',
    'krfr',
    'krpl',
    'krst',
    'kres',
    'ksav',
    'kspd',
    'kund',
    'kBEG',
    'kCAN',
    'kCMD',
    'kCPY',
    'kCRT',
    'kDC',
    'kDL',
    'kslt',
    'kEND',
    'kEOL',
    'kEXT',
    'kFND',
    'kHLP',
    'kHOM',
    'kIC',
    'kLFT',
    'kMSG',
    'kMOV',
    'kNXT',
    'kOPT',
    'kPRV',
    'kPRT',
    'kRDO',
    'kRPL',
    'kRIT',
    'kRES',
    'kSAV',
    'kSPD',
    'kUND',
    'rfi',
    'kf11',
    'kf12',
    'kf13',
    'kf14',
    'kf15',
    'kf16',
    'kf17',
    'kf18',
    'kf19',
    'kf20',
    'kf21',
    'kf22',
    'kf23',
    'kf24',
    'kf25',
    'kf26',
    'kf27',
    'kf28',
    'kf29',
    'kf30',
    'kf31',
    'kf32',
    'kf33',
    'kf34',
    'kf35',
    'kf36',
    'kf37',
    'kf38',
    'kf39',
    'kf40',
    'kf41',
    'kf42',
    'kf43',
    'kf44',
    'kf45',
    'kf46',
    'kf47',
    'kf48',
    'kf49',
    'kf50',
    'kf51',
    'kf52',
    'kf53',
    'kf54',
    'kf55',
    'kf56',
    'kf57',
    'kf58',
    'kf59',
    'kf60',
    'kf61',
    'kf62',
    'kf63',
    'el1',
    'mgc',
    'smgl',
    'smgr',
    'fln',
    'sclk',
    'dclk',
    'rmclk',
    'cwin',
    'wingo',
    'hup',
    'dial',
    'qdial',
    'tone',
    'pulse',
    'hook',
    'pause',
    'wait',
    'u0',
    'u1',
    'u2',
    'u3',
    'u4',
    'u5',
    'u6',
    'u7',
    'u8',
    'u9',
    'op',
    'oc',
    'initc',
    'initp',
    'scp',
    'setf',
    'setb',
    'cpi',
    'lpi',
    'chr',
    'cvr',
    'defc',
    'swidm',
    'sdrfq',
    'sitm',
    'slm',
    'smicm',
    'snlq',
    'snrmq',
    'sshm',
    'ssubm',
    'ssupm',
    'sum',
    'rwidm',
    'ritm',
    'rlm',
    'rmicm',
    'rshm',
    'rsubm',
    'rsupm',
    'rum',
    'mhpa',
    'mcud1',
    'mcub1',
    'mcuf1',
    'mvpa',
    'mcuu1',
    'porder',
    'mcud',
    'mcub',
    'mcuf',
    'mcuu',
    'scs',
    'smgb',
    'smgbp',
    'smglp',
    'smgrp',
    'smgt',
    'smgtp',
    'sbim',
    'scsd',
    'rbim',
    'rcsd',
    'subcs',
    'supcs',
    'docr',
    'zerom',
    'csnm',
    'kmous',
    'minfo',
    'reqmp',
    'getm',
    'setaf',
    'setab',
    'pfxl',
    'devt',
    'csin',
    's0ds',
    's1ds',
    's2ds',
    's3ds',
    'smglr',
    'smgtb',
    'birep',
    'binel',
    'bicr',
    'colornm',
    'defbi',
    'endbi',
    'setcolor',
    'slines',
    'dispc',
    'smpch',
    'rmpch',
    'smsc',
    'rmsc',
    'pctrm',
    'scesc',
    'scesa',
    'ehhlm',
    'elhlm',
    'elohlm',
    'erhlm',
    'ethlm',
    'evhlm',
    'sgr1',
    'slength',
    'OTi2',
    'OTrs',
    'OTnl',
    'OTbc',
    'OTko',
    'OTma',
    'OTG2',
    'OTG3',
    'OTG1',
    'OTG4',
    'OTGR',
    'OTGL',
    'OTGU',
    'OTGD',
    'OTGH',
    'OTGV',
    'OTGC',
    'meml',
    'memu',
    'box1',
)


def _get_terminfo_dirs() -> list[pathlib.Path]:
    """
    Get list of directories to search for terminfo files.

    Based on ncurses behavior in:
    - ncurses/tinfo/db_iterator.c:_nc_next_db()
    - ncurses/tinfo/read_entry.c:_nc_read_entry()
    """

    dirs = []

    terminfo = os.environ.get('TERMINFO')
    if terminfo:
        dirs.append(terminfo)

    try:
        home = pathlib.Path.home()
        dirs.append(str(home / '.terminfo'))
    except RuntimeError:
        pass

    # Check TERMINFO_DIRS
    terminfo_dirs = os.environ.get('TERMINFO_DIRS', '')
    if terminfo_dirs:
        for d in terminfo_dirs.split(':'):
            if d:
                dirs.append(d)

    dirs.extend([
        '/etc/terminfo',
        '/lib/terminfo',
        '/usr/lib/terminfo',
        '/usr/share/terminfo',
        '/usr/share/lib/terminfo',
        '/usr/share/misc/terminfo',
        '/usr/local/lib/terminfo',
        '/usr/local/share/terminfo',
    ])

    return [pathlib.Path(d) for d in dirs if pathlib.Path(d).is_dir()]


def _validate_terminal_name_or_raise(terminal_name: str) -> None:
    if not isinstance(terminal_name, str):
        raise TypeError('`terminal_name` must be a string')

    if not terminal_name:
        raise ValueError('`terminal_name` cannot be empty')

    if '\x00' in terminal_name:
        raise ValueError('NUL character found in `terminal_name`')

    t = pathlib.Path(terminal_name)
    if len(t.parts) > 1:
        raise ValueError('`terminal_name` cannot contain path separators')


def _read_terminfo_file(terminal_name: str) -> bytes:
    """
    Find and read terminfo file for given terminal name.

    Terminfo files are stored in directories using the first character of the terminal name as a subdirectory.
    """

    _validate_terminal_name_or_raise(terminal_name)
    first_char = terminal_name[0].lower()
    filename = terminal_name

    for directory in _get_terminfo_dirs():
        path = directory / first_char / filename
        if path.is_file():
            return path.read_bytes()

        # Try with hex encoding of first char (for special chars)
        hex_dir = f'{ord(first_char):02x}'
        path = directory / hex_dir / filename
        if path.is_file():
            return path.read_bytes()

    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)


def _parse_terminfo_file(terminal_name: str) -> ta.Mapping[str, bytes]:
    """
    Parse a terminfo file.

    Populate the _capabilities dict for easy retrieval

    Based on ncurses implementation in:
    - ncurses/tinfo/read_entry.c:_nc_read_termtype()
    - ncurses/tinfo/read_entry.c:_nc_read_file_entry()
    - ncurses/tinfo/lib_ti.c:tigetstr()
    """

    data = _read_terminfo_file(terminal_name)
    too_short = f'TermInfo file for {terminal_name!r} too short'
    offset = 12
    if len(data) < offset:
        raise ValueError(too_short)

    magic, name_size, bool_count, num_count, str_count, str_size = \
        struct.unpack('<Hhhhhh', data[:offset])

    if magic == _MAGIC16:
        number_size = 2
    elif magic == _MAGIC32:
        number_size = 4
    else:
        raise ValueError(f'TermInfo file for {terminal_name!r} uses unknown magic')

    # Skip data than PyREPL doesn't need:
    # - names (`|`-separated ASCII strings)
    # - boolean capabilities (bytes with value 0 or 1)
    # - numbers (little-endian integers, `number_size` bytes each)
    offset += name_size
    offset += bool_count
    if offset % 2:
        # Align to even byte boundary for numbers
        offset += 1
    offset += num_count * number_size
    if offset > len(data):
        raise ValueError(too_short)

    # Read string offsets
    end_offset = offset + 2 * str_count
    if offset > len(data):
        raise ValueError(too_short)
    string_offset_data = data[offset:end_offset]
    string_offsets = [off for [off] in struct.iter_unpack('<h', string_offset_data)]
    offset = end_offset

    # Read string table
    if offset + str_size > len(data):
        raise ValueError(too_short)
    string_table = data[offset: offset + str_size]

    # Extract strings from string table
    capabilities = {}
    for cap, off in zip(_ORDERED_STRING_NAMES, string_offsets):
        if off < 0:
            # CANCELLED_STRING; we do not store those
            continue
        elif off < len(string_table):
            # Find null terminator
            end = string_table.find(0, off)
            if end >= 0:
                capabilities[cap] = string_table[off:end]
        # in other cases this is ABSENT_STRING; we don't store those.

    # Note: we don't support extended capabilities since PyREPL doesn't need them.

    return capabilities


##


@dc.dataclass(frozen=True)
class TermInfo:
    terminal_name: str
    capabilities: ta.Mapping[str, bytes]

    is_fallback: bool = False

    def get(self, cap: str) -> bytes | None:
        """Get terminal capability string by name."""

        if not isinstance(cap, str):
            raise TypeError(f'`cap` must be a string, not {type(cap)}')

        return self.capabilities.get(cap)


def load_term_info(
        terminal_name: str | bytes | None,
        *,
        fallback: bool = True,
) -> TermInfo:
    """
    Initialize terminal capabilities for the given terminal type.

    Based on ncurses implementation in:
    - ncurses/tinfo/lib_setup.c:setupterm() and _nc_setupterm()
    - ncurses/tinfo/lib_setup.c:TINFO_SETUP_TERM()

    This version first attempts to read terminfo database files like ncurses, then, if `fallback` is True, falls
    back to hardcoded capabilities for common terminal types.
    """

    # If termstr is None or empty, try to get from environment
    if not terminal_name:
        terminal_name = os.environ.get('TERM') or 'ANSI'

    if isinstance(terminal_name, bytes):
        terminal_name = terminal_name.decode('ascii')

    is_fallback = False
    try:
        capabilities = _parse_terminfo_file(terminal_name)
    except (OSError, ValueError):
        if not fallback:
            raise

        term_type = _TERM_ALIASES.get(terminal_name, terminal_name)
        if term_type not in _FALLBACK_TERMINAL_CAPABILITIES:
            term_type = 'dumb'

        is_fallback = True
        capabilities = _FALLBACK_TERMINAL_CAPABILITIES[term_type]

    return TermInfo(
        terminal_name=terminal_name,
        capabilities=capabilities,
        is_fallback=is_fallback,
    )


##


def tparm(cap_bytes: bytes, *params: int) -> bytes:
    """
    Parameterize a terminal capability string.

    Based on ncurses implementation in:
    - ncurses/tinfo/lib_tparm.c:tparm()
    - ncurses/tinfo/lib_tparm.c:tparam_internal()

    The ncurses version implements a full stack-based interpreter for terminfo parameter strings. This pure Python
    version implements only the subset of parameter substitution operations needed by PyREPL:
    - %i (increment parameters for 1-based indexing)
    - %p[1-9]%d (parameter substitution)
    - %p[1-9]%{n}%+%d (parameter plus constant)
    """

    if not isinstance(cap_bytes, bytes):
        raise TypeError(f'`cap` must be bytes, not {type(cap_bytes)}')

    result = cap_bytes

    # %i - increment parameters (1-based instead of 0-based)
    increment = b'%i' in result
    if increment:
        result = result.replace(b'%i', b'')

    # Replace %p1%d, %p2%d, etc. with actual parameter values
    for i in range(len(params)):
        pattern = b'%%p%d%%d' % (i + 1)
        if pattern in result:
            value = params[i]
            if increment:
                value += 1
            result = result.replace(pattern, str(value).encode('ascii'))

    # Handle %p1%{1}%+%d (parameter plus constant)
    # Used in some cursor positioning sequences
    pattern_re = re.compile(rb'%p(\d)%\{(\d+)\}%\+%d')
    matches = list(pattern_re.finditer(result))
    for match in reversed(matches):  # reversed to maintain positions
        param_idx = int(match.group(1))
        constant = int(match.group(2))
        value = params[param_idx] + constant
        result = (
            result[: match.start()] +
            str(value).encode('ascii') +
            result[match.end():]
        )

    return result



"""
TODO:
 - standardize following redirects
"""
import contextlib
import functools
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ...io.streams.adapters import ByteStreamBufferAsyncBytesReaderAdapter
from ...io.streams.adapters import ByteStreamBufferBytesReaderAdapter
from ...io.streams.segmented import SegmentedByteStreamBuffer
from ..headers import HttpHeaders
from .asyncs import AsyncHttpClient
from .asyncs import AsyncStreamHttpClientResponse
from .base import HttpClientContext
from .base import HttpClientError
from .base import HttpClientRequest
from .sync import HttpClient
from .sync import StreamHttpClientResponse


if ta.TYPE_CHECKING:
    import httpx
else:
    httpx = lang.proxy_import('httpx')


##


class HttpxHttpClient(HttpClient):
    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        it: ta.Iterator[bytes]

        def read1(self, n: int = -1, /) -> bytes:
            try:
                return next(self.it)
            except StopIteration:
                return b''

    def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> StreamHttpClientResponse:
        try:
            resp_cm = httpx.stream(
                method=req.method_or_default,
                url=req.url,
                headers=req.headers_.all if req.headers_ is not None else None,
                content=req.data,
                timeout=req.timeout_s,
            )

        except httpx.HTTPError as e:
            raise HttpClientError from e

        resp_close = functools.partial(resp_cm.__exit__, None, None, None)

        try:
            resp = resp_cm.__enter__()
            return StreamHttpClientResponse(
                status=resp.status_code,
                headers=HttpHeaders(resp.headers.raw),
                request=req,
                underlying=resp,
                _stream=ByteStreamBufferBytesReaderAdapter.wrap(
                    self._StreamAdapter(resp.iter_bytes()),
                    SegmentedByteStreamBuffer(chunk_size=16 * 1024),
                ),
                _closer=resp_close,  # type: ignore
            )

        except httpx.HTTPError as e:
            resp_close()
            raise HttpClientError from e

        except BaseException:
            resp_close()
            raise


##


class HttpxAsyncHttpClient(AsyncHttpClient):
    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        it: ta.AsyncIterator[bytes]

        async def read1(self, n: int = -1, /) -> bytes:
            try:
                return await anext(self.it)
            except StopAsyncIteration:
                return b''

    async def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> AsyncStreamHttpClientResponse:
        es = contextlib.AsyncExitStack()

        try:
            client = await es.enter_async_context(httpx.AsyncClient())

            resp = await es.enter_async_context(client.stream(
                method=req.method_or_default,
                url=req.url,
                headers=req.headers_.all if req.headers_ is not None else None,
                content=req.data,
                timeout=req.timeout_s,
            ))

            it = resp.aiter_bytes()

            # FIXME:
            #  this has a tendency to raise `RuntimeError: async generator ignored GeneratorExit` when all of the
            #  following conditions are met:
            #   - stopped iterating midway through
            #   - shutting down the event loop
            #   - debugging under pycharm / pydevd
            #   - running under asyncio
            #  it does not seem to happen unless all of these conditions are met. see:
            #   https://gist.github.com/wrmsr/a0578ee5d5371b53804cfb56aeb84cdf .
            es.push_async_callback(it.aclose)  # type: ignore[attr-defined]

            return AsyncStreamHttpClientResponse(
                status=resp.status_code,
                headers=HttpHeaders(resp.headers.raw),
                request=req,
                underlying=resp,
                _stream=ByteStreamBufferAsyncBytesReaderAdapter.wrap(
                    self._StreamAdapter(it),
                    SegmentedByteStreamBuffer(chunk_size=16 * 1024),
                ),
                _closer=es.aclose,
            )

        except httpx.HTTPError as e:
            await es.aclose()
            raise HttpClientError from e

        except BaseException:
            await es.aclose()
            raise



import sys
import typing as ta


##


def confirm_action(
        message: str | None = None,
        *,
        stdin: ta.Any | None = None,
        stdout: ta.Any | None = None,
) -> bool:
    if stdin is None:
        stdin = sys.stdin
    if not stdin.isatty():
        raise OSError(f'stdin {stdin!r} is not a tty')
    # FIXME: we want to make sure we only run on a tty, but we also want input()'s readline goodies..
    if stdin is not sys.stdin:
        raise RuntimeError('Unsupported stdin')

    if stdout is None:
        stdout = sys.stdout
    if not stdout.isatty():
        raise OSError(f'stdout {stdout!r} is not a tty')

    if message:
        if '\n' in message:
            prefix = message + '\n\n'
        else:
            prefix = message + ' '
    else:
        prefix = ''

    while True:
        c = input(f'{prefix}(y/n): ').lower().strip()

        if c == 'y':
            return True
        elif c == 'n':
            return False
        else:
            print("Please enter 'y' for yes or 'n' for no.", file=stdout)



# @omlish-lite
import asyncio
import typing as ta

from ..identities import AsyncliteIdentities


##


class AsyncioAsyncliteIdentities(AsyncliteIdentities):
    def current_identity(self) -> ta.Any:
        return asyncio.current_task()



from .api import (  # noqa
    Api,
)

from .app import (  # noqa
    AppRunParams,
    AppRunner,

    ViewFunc,
    BeforeRequestFunc,
    AfterRequestFunc,
    App,
)

from .cvs import (  # noqa
    CvLookupError,

    Cvs,
)

from .requests import (  # noqa
    Request,
)

from .responses import (  # noqa
    ResponseData,
    ResponseStatus,
    ResponseHeaders,

    Response,
)

from .routes import (  # noqa
    RouteKey,
    Route,
)

from .types import (  # noqa
    ImmutableMultiDict,
)



# ruff: noqa: FURB188 UP006 UP045
# @omlish-lite
"""
TODO:
 - netty 'pending_removal' trick
"""
import abc
import typing as ta

from ....lite.abstract import Abstract
from ....lite.check import check
from ...streams.direct import DirectByteStreamBuffer
from ...streams.framing import LongestMatchDelimiterByteStreamFrameDecoder
from ...streams.scanning import ScanningByteStreamBuffer
from ...streams.segmented import SegmentedByteStreamBuffer
from ...streams.types import ByteStreamBuffer
from ...streams.types import MutableByteStreamBuffer
from ...streams.utils import ByteStreamBuffers
from ...streams.utils import CanByteStreamBuffer
from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineMessages
from ..errors import IncompleteDecodingIoPipelineError
from ..flow.types import IoPipelineFlow
from ..flow.types import IoPipelineFlowMessages
from .buffering import InboundBytesBufferingIoPipelineHandler


##


class UnicodeDecoderIoPipelineHandler(IoPipelineHandler):
    """bytes/view -> str (UTF-8, replacement)."""

    def __init__(
            self,
            encoding: str = 'utf-8',
            *,
            errors: ta.Literal['strict', 'ignore', 'replace'] = 'strict',
    ) -> None:
        super().__init__()

        self._encoding = encoding
        self._errors = errors

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if ByteStreamBuffers.can_bytes(msg):
            b = ByteStreamBuffers.to_bytes(msg)

            msg = b.decode(self._encoding, errors=self._errors)

        ctx.feed_in(msg)


##


class DelimiterFrameDecoderIoPipelineHandler(InboundBytesBufferingIoPipelineHandler):
    """
    bytes-like -> frames using longest-match delimiter semantics.

    TODO:
     - flow control, *or* replace with BytesToMessageDecoderIoPipelineHandler
    """

    def __init__(
            self,
            delims: ta.Sequence[bytes],
            *,
            keep_ends: bool = False,
            max_size: ta.Optional[int] = None,
            max_buffer: ta.Optional[int] = None,
            buffer_chunk_size: int = 64 * 1024,
            on_incomplete_final: ta.Literal['allow', 'raise'] = 'allow',
    ) -> None:
        super().__init__()

        self._on_incomplete_final = on_incomplete_final

        self._buf = ScanningByteStreamBuffer(SegmentedByteStreamBuffer(
            max_size=max_buffer,
            chunk_size=buffer_chunk_size,
        ))

        self._fr = LongestMatchDelimiterByteStreamFrameDecoder(
            delims,
            keep_ends=keep_ends,
            max_size=max_size,
        )

    def inbound_buffered_bytes(self) -> int:
        return len(self._buf)

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineMessages.FinalInput):
            self._produce_frames(ctx, final=True)
            ctx.feed_in(msg)
            return

        if not ByteStreamBuffers.can_bytes(msg):
            ctx.feed_in(msg)
            return

        for mv in ByteStreamBuffers.iter_segments(msg):
            if mv:
                self._buf.write(mv)

        self._produce_frames(ctx)

    def _produce_frames(self, ctx: IoPipelineHandlerContext, *, final: bool = False) -> None:
        frames = self._fr.decode(self._buf, final=final)

        if final and len(self._buf):
            if (oif := self._on_incomplete_final) == 'allow':
                frames.append(self._buf.split_to(len(self._buf)))
            elif oif == 'raise':
                raise IncompleteDecodingIoPipelineError
            else:
                raise RuntimeError(f'unexpected on_incomplete_final: {oif!r}')

        for fr in frames:
            ctx.feed_in(fr)


##


class BytesToMessageDecoderIoPipelineHandler(IoPipelineHandler, Abstract):
    @abc.abstractmethod
    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            data: CanByteStreamBuffer,
            out: ta.List[ta.Any],
            *,
            final: bool = False,
    ) -> None:
        raise NotImplementedError

    #

    _called_decode: bool = False  # ~ `selfFiredChannelRead`
    _produced_messages: bool = False  # ~ `firedChannelRead`

    def _call_decode(
            self,
            ctx: IoPipelineHandlerContext,
            data: CanByteStreamBuffer,
            *,
            final: bool = False,
    ) -> None:
        self._called_decode = True

        out: ta.List[ta.Any] = []
        self._decode(ctx, data, out, final=final)

        if not out:
            return

        self._produced_messages = True

        for out_msg in out:
            ctx.feed_in(out_msg)

    #

    def _on_bytes_input(self, ctx: IoPipelineHandlerContext, data: CanByteStreamBuffer) -> None:
        check.arg(len(data) > 0)

        self._call_decode(ctx, data)

    def _on_flush_input(self, ctx: IoPipelineHandlerContext) -> None:
        if (
                self._called_decode and
                not self._produced_messages and
                not ctx.services[IoPipelineFlow].is_auto_read()
        ):
            ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())

        self._called_decode = False
        self._produced_messages = False

        ctx.feed_in(IoPipelineFlowMessages.FlushInput())

    def _on_final_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineMessages.FinalInput) -> None:
        self._call_decode(ctx, DirectByteStreamBuffer(b''), final=True)

        ctx.feed_in(msg)

    #

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineFlowMessages.FlushInput):
            self._on_flush_input(ctx)

        elif isinstance(msg, IoPipelineMessages.FinalInput):
            self._on_final_input(ctx, msg)

        elif ByteStreamBuffers.can_bytes(msg):
            self._on_bytes_input(ctx, msg)

        else:
            ctx.feed_in(msg)


#


@ta.final
class FnBytesToMessageDecoderIoPipelineHandler(BytesToMessageDecoderIoPipelineHandler):
    class DecodeFn(ta.Protocol):
        def __call__(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> None:
            ...

    def __init__(
            self,
            decode_fn: DecodeFn,
    ) -> None:
        super().__init__()

        self._decode_fn = decode_fn

    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            buf: CanByteStreamBuffer,
            out: ta.List[ta.Any],
            *,
            final: bool = False,
    ) -> None:
        self._decode_fn(ctx, buf, out, final=final)


##


class BufferedBytesToMessageDecoderIoPipelineHandler(
    InboundBytesBufferingIoPipelineHandler,
    BytesToMessageDecoderIoPipelineHandler,
    Abstract,
):
    def __init__(
            self,
            *,
            max_buffer_size: ta.Optional[int] = None,
            buffer_chunk_size: int = 64 * 1024,
            scanning_buffer: bool = False,
    ) -> None:
        super().__init__()

        self._max_buffer_size = max_buffer_size
        self._buffer_chunk_size = buffer_chunk_size
        self._scanning_buffer = scanning_buffer

    #

    def inbound_buffered_bytes(self) -> int:
        if (buf := self._buf) is None:
            return 0
        return len(buf)

    _buf: ta.Optional[MutableByteStreamBuffer] = None

    def _new_buf(self) -> MutableByteStreamBuffer:
        buf: MutableByteStreamBuffer = SegmentedByteStreamBuffer(
            max_size=self._max_buffer_size,
            chunk_size=self._buffer_chunk_size,
        )

        if self._scanning_buffer:
            buf = ScanningByteStreamBuffer(buf)

        return buf

    #

    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            data: CanByteStreamBuffer,
            out: ta.List[ta.Any],
            *,
            final: bool = False,
    ) -> None:
        if final:
            check.arg(len(data) == 0)

            if not isinstance(data, ByteStreamBuffer):
                data = DirectByteStreamBuffer(b'')

            self._decode_buffer(ctx, data, out, final=final)

            return

        check.arg(len(data) > 0)

        if (buf := self._buf) is None:
            buf = self._buf = self._new_buf()

        for seg in ByteStreamBuffers.iter_segments(data):
            buf.write(seg)

        self._decode_buffer(ctx, buf, out, final=final)

    #

    @abc.abstractmethod
    def _decode_buffer(
            self,
            ctx: IoPipelineHandlerContext,
            buf: ByteStreamBuffer,
            out: ta.List[ta.Any],
            *,
            final: bool = False,
    ) -> None:
        raise NotImplementedError



"""
TODO:
 - Configurable? class Config crtp?
  - all config on specs? prob not desirable purity, prob want ~some~ global config if only defaults
   - ProcessorConfigSpecProviderFactoryThingy?
  - FIXME: for now these are processor ctor kwargs that are never overridden
"""
import typing as ta

from .... import check
from .... import lang
from ...specs import ClassSpec
from ..configs import DEFAULT_NAMED_PACKAGE_CONFIG
from ..configs import NamedPackageConfig


T = ta.TypeVar('T')
ProcessingOptionT = ta.TypeVar('ProcessingOptionT', bound='ProcessingOption')


##


ProcessingContextItemFactory: ta.TypeAlias = ta.Callable[['ProcessingContext'], ta.Any]


class ProcessingOption(lang.Abstract):
    pass


class ProcessingContext:
    def __init__(
            self,
            cls: type,
            cs: ClassSpec,
            item_factories: ta.Mapping[type, ProcessingContextItemFactory],
            *,
            pkg_cfg: NamedPackageConfig = DEFAULT_NAMED_PACKAGE_CONFIG,
            options: ta.Sequence[ProcessingOption] | None = None,
    ) -> None:
        super().__init__()

        self._cls = cls
        self._cs = cs
        self._item_factories = item_factories
        self._pkg_cfg = pkg_cfg

        options_dct: dict = {}
        for o in options or ():
            check.not_in(type(o), options_dct)
            options_dct[type(o)] = o
        self._options_dct = options_dct

        self._items: dict = {}

    @property
    def cls(self) -> type:
        return self._cls

    @property
    def cs(self) -> ClassSpec:
        return self._cs

    @property
    def pkg_cfg(self) -> NamedPackageConfig:
        return self._pkg_cfg

    def __getitem__(self, ty: type[T]) -> T:
        try:
            return self._items[ty]
        except KeyError:
            pass

        fac = self._item_factories[ty]
        ret = fac(self)
        self._items[ty] = ret
        return ret

    def option(self, ty: type[ProcessingOptionT]) -> ProcessingOptionT | None:
        return self._options_dct.get(ty)


##


class Processor(lang.Abstract):
    def __init__(self, ctx: ProcessingContext) -> None:
        super().__init__()

        self._ctx = ctx

    def check(self) -> None:
        pass

    def process(self, cls: type) -> type:
        return cls



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from ....lite.json import json_dumps_pretty
from ..backends import YamlBackend
from .decoding import yaml_decode


##


class GoyamlYamlBackend(YamlBackend):
    def is_available(self) -> bool:
        return True

    def loads(self, s: str) -> ta.Any:
        return yaml_decode(s)

    def dumps(self, o: ta.Any) -> str:
        return json_dumps_pretty(o)



from ... import dataclasses as _dc  # noqa


_dc.init_package(
    globals(),
    codegen=True,
)


##


from .openapi import (  # noqa
    Openapi,
)


##


from ... import marshal as _msh  # noqa

_msh.register_global_module_import('._marshal', __package__)
